import numpy as np
from pydantic import field_validator, confloat, Field, AliasChoices
from typing import List, Type, Union, Dict

from ._functions import _rotation_matrix

from .baseModels import IgnoreExtra, NumpyVectorModel, T


class Position(NumpyVectorModel):
    """
    Position model. Cartesian co-ordinates are used.
    """

    x: float = 0.0
    """Horizontal position [m]."""

    y: float = 0.0
    """Vertical position [m]."""

    z: float = 0.0
    """Longitudinal position [m]."""

    def __add__(self, other: Type[T]) -> T:
        return Position(
            x=(self.x + other.x), y=(self.y + other.y), z=(self.z + other.z)
        )

    def __radd__(self, other: Type[T]) -> T:
        return self.__add__(other)

    def __sub__(self, other: Type[T]) -> T:
        return Position(
            x=(self.x - other.x), y=(self.y - other.y), z=(self.z - other.z)
        )

    def __rsub__(self, other: Type[T]) -> T:
        return Position(
            x=(other.x - self.x), y=(other.y - self.y), z=(other.z - self.z)
        )

    def dot(self, other: Union[List, Type[T]]) -> float:
        if isinstance(other, (set, tuple, list)):
            other = Position.from_list(other)
        return self.x * other.x + self.y * other.y + self.z * other.z

    def vector_angle(self, other: Union[List, Type[T]], direction: List) -> float:
        if isinstance(other, (set, tuple, list)):
            other = Position.from_list(other)
        return (self - other).dot(direction)

    def length(self) -> float:
        return np.sqrt([self.x * self.x + self.y * self.y + self.z * self.z])


class Rotation(NumpyVectorModel):
    """
    Rotation model.
    """

    phi: confloat(ge=-np.pi, le=np.pi) = 0.0  # type: ignore
    """Rotation about the horizontal axis [rad]."""

    psi: confloat(ge=-np.pi, le=np.pi) = 0.0  # type: ignore
    """Rotation about the vertical axis [rad]."""

    theta: confloat(ge=-np.pi, le=np.pi) = 0.0  # type: ignore
    """Rotation about the longitudinal axis [rad]."""

    def __add__(self, other: Type[T]) -> T:
        return Rotation(
            phi=(self.phi + other.phi),
            psi=(self.psi + other.psi),
            theta=(self.theta + other.theta),
        )

    def __radd__(self, other: Type[T]) -> T:
        return self.__add__(other)

    def __sub__(self, other: Type[T]) -> T:
        return Rotation(
            phi=(self.phi - other.phi),
            psi=(self.psi - other.psi),
            theta=(self.theta - other.theta),
        )

    def __rsub__(self, other: Type[T]) -> T:
        return Rotation(
            phi=(other.phi - self.phi),
            psi=(other.psi - self.psi),
            theta=(other.theta - self.theta),
        )

    def __abs__(self):
        return Rotation(phi=abs(self.phi), psi=abs(self.psi), theta=abs(self.theta))

    def __gt__(self, value: Union[int, float, List, Type[T]]):
        if isinstance(value, (int, float)):
            return any([self.phi > value, self.psi > value, self.theta > value])
        elif isinstance(value, (Union[list, set, tuple])):
            return [self.phi, self.psi, self.theta] > value
        elif isinstance(value, Rotation):
            return any(
                [self.phi > value.phi, self.psi > value.psi, self.theta > value.theta]
            )


class ElementError(IgnoreExtra):
    """
    Position/Rotation error model.
    """

    position: Union[Position, List[Union[float, int]]] = Position(x=0, y=0, z=0)
    """Errors in position."""

    rotation: Union[Rotation, List[Union[float, int]]] = Rotation(theta=0, phi=0, psi=0)
    """Errors in rotation."""

    @field_validator("position", mode="before")
    @classmethod
    def validate_position(cls, v: Union[Position, Dict, List, np.ndarray]) -> Position:
        if isinstance(v, (list, tuple, np.ndarray)) and len(v) == 3:
            return Position(x=v[0], y=v[1], z=v[2])
        elif isinstance(v, Position):
            return v
        elif isinstance(v, dict):
            keys = list(v.keys())
            values = list(v.values())
            if all([x in keys for x in ["x", "y", "z"]]) and all([type(val) == float for val in values]) and len(
                    keys) == 3:
                return Position(**v)
            else:
                raise ValueError("setting middle as dictionary must include x, y, z as floats")

        else:
            raise ValueError("position should be a number or a list of floats")

    @field_validator("rotation", mode="before")
    @classmethod
    def validate_rotation(cls, v: Union[Rotation, Dict, List, np.ndarray]) -> Rotation:
        if isinstance(v, (list, tuple, np.ndarray)) and len(v) == 3:
            return Rotation(theta=v[0], phi=v[1], psi=v[2])
        elif isinstance(v, Rotation):
            return v
        elif isinstance(v, dict):
            keys = list(v.keys())
            values = list(v.values())
            if all([x in keys for x in ["phi", "psi", "theta"]]) and all(
                    [type(val) == float for val in values]) and len(keys) == 3:
                return Rotation(**v)
            else:
                raise ValueError("setting rotation as dictionary must include x, y, z as floats")

        else:
            raise ValueError("rotation should be a number or a list of floats")

    def __str__(self):
        cls = self.__class__
        if any([getattr(self, k) != 0 for k in cls.model_fields]):
            return " ".join(
                [
                    getattr(self, k).__repr__()
                    for k in cls.model_fields
                    if getattr(self, k) != 0
                ]
            )
        else:
            return str(None)

    def __repr__(self):
        return self.__class__.__name__ + "(" + self.__str__() + ")"

    def __eq__(self, other):
        cls = self.__class__
        if other == 0:
            return all([getattr(self, k) == 0 for k in cls.model_fields.keys()])
        else:
            return super().__eq__(other)


class ElementSurvey(ElementError):
    pass


class PhysicalElement(IgnoreExtra):
    """
    Physical info model.
    """

    middle: Position = Field(default=Position(), alias=AliasChoices("position", "centre"))
    """Middle position of the element."""

    datum: Position = Field(default=0)
    """Datum."""

    rotation: Rotation = Rotation(theta=0, phi=0, psi=0)
    """Local rotation of the element."""

    global_rotation: Rotation = Rotation(theta=0, phi=0, psi=0)
    """Global rotation of the element."""

    error: ElementError = ElementError()
    """Position errors in the element."""

    survey: ElementSurvey = ElementSurvey()
    """Survey positions of the element."""

    length: float = 0.0
    """Length of the element."""

    physical_angle: float = 0.0
    """Physical angle"""

    def __str__(self):
        cls = self.__class__
        if any([getattr(self, k) != 0 for k in cls.model_fields.keys()]):
            return " ".join(
                [
                    str(k) + "=" + getattr(self, k).__repr__()
                    for k in cls.model_fields.keys()
                    if getattr(self, k) != 0
                ]
            )
        else:
            return str()

    def __repr__(self):
        return self.__class__.__name__ + "(" + self.__str__() + ")"

    @field_validator("middle", "datum", mode="before")
    @classmethod
    def validate_middle(cls, v: Union[float, int, Dict, List, np.ndarray]) -> Position:
        if isinstance(v, (float, int)):
            return Position(z=v)
        elif isinstance(v, (list, tuple, np.ndarray)):
            if len(v) == 3:
                return Position(x=v[0], y=v[1], z=v[2])
            elif len(v) == 2:
                return Position(x=v[0], y=0, z=v[1])
        elif isinstance(v, Position):
            return v
        elif isinstance(v, dict):
            keys = list(v.keys())
            values = list(v.values())
            if all([x in keys for x in ["x", "y", "z"]]) and all([type(val) == float for val in values]) and len(
                    keys) == 3:
                return Position(**v)
            else:
                raise ValueError("setting middle as dictionary must include x, y, z as floats")

        else:
            raise ValueError("middle should be a number or a list of floats")

    @field_validator("rotation", "global_rotation", mode="before")
    @classmethod
    def validate_rotation(cls, v: Union[float, int, List, np.ndarray]) -> Rotation:
        if isinstance(v, (float, int)):
            return Rotation(theta=v)
        elif isinstance(v, (list, tuple, np.ndarray)):
            if len(v) == 3:
                return Rotation(phi=v[0], psi=v[1], theta=v[2])
        elif isinstance(v, Rotation):
            return v
        elif isinstance(v, dict):
            keys = list(v.keys())
            values = list(v.values())
            if all([x in keys for x in ["phi", "psi", "theta"]]) and all(
                    [type(val) == float for val in values]) and len(keys) == 3:
                return Rotation(**v)
            else:
                raise ValueError("setting rotation as dictionary must include x, y, z as floats")

        else:
            raise ValueError("rotation should be a number or a list of floats")

    @property
    def rotation_matrix(self) -> np.ndarray:
        """
        Get the 3D rotation matrix of the element based on combined rotations.
        Convention: [pitch, roll, yaw] = [rotation around X, rotation around Z, rotation around Y]

        Returns
        -------
        np.ndarray
            3x3 Rotation matrix
        """
        # Get the combined rotation angles
        pitch = self.rotation.phi + self.global_rotation.phi  # X rotation (pitch) - affects Y,Z
        roll = self.rotation.psi + self.global_rotation.psi  # Z rotation (roll) - affects X,Y
        yaw = self.rotation.theta + self.global_rotation.theta  # Y rotation (yaw) - affects X,Z

        # Rotation matrix around X axis (pitch)
        Rx = np.array([
            [1, 0, 0],
            [0, np.cos(pitch), -np.sin(pitch)],
            [0, np.sin(pitch), np.cos(pitch)]
        ])

        # Rotation matrix around Z axis (roll)
        Rz = np.array([
            [np.cos(roll), -np.sin(roll), 0],
            [np.sin(roll), np.cos(roll), 0],
            [0, 0, 1]
        ])

        # Rotation matrix around Y axis (yaw) - this is the horizontal bending
        Ry = np.array([
            [np.cos(yaw), 0, np.sin(yaw)],
            [0, 1, 0],
            [-np.sin(yaw), 0, np.cos(yaw)]
        ])

        # Combined rotation matrix - apply yaw first (most common), then pitch, then roll
        return Rx @ Rz @ Ry

    def rotated_position(
            self, vec: List[Union[int, float]] = [0, 0, 0]
    ) -> np.ndarray:
        """
        Get the rotated position of the element based on matrix multiplication with rotation_matrix.

        Parameters
        ----------
        vec: List[float]
            Vector by which to rotate the element

        Returns
        -------
        np.ndarray
            Rotated vector.
        """
        return self.rotation_matrix @ np.array(vec)

    @property
    def start(self) -> Position:
        """
        Start position of the element.

        Returns
        -------
        :class:`~nala.models.physical.Position
            Start position of the element.
        """
        middle = np.array(self.middle.array)

        # Calculate local offset from middle to start
        if abs(self.physical_angle) > 1e-9:
            # Bent element - correct arc geometry
            # The start is offset from middle by half the chord deviation
            sx = -self.length * (1 - np.cos(self.physical_angle)) / (2 * self.physical_angle)
            sy = 0
            sz = -self.length * np.sin(self.physical_angle) / (2 * self.physical_angle)
        else:
            # Straight element
            sx = 0
            sy = 0
            sz = -self.length / 2.0

        # Local offset vector
        vec = [sx, sy, sz]

        # Rotate to global coordinates and add to middle
        start = middle + self.rotated_position(vec)
        return Position.from_list(start)

    @property
    def end(self) -> Position:
        """
        End position of the element.

        Returns
        -------
        :class:`~nala.models.physical.Position
            End position of the element.
        """
        middle = np.array(self.middle.array)

        # Calculate local offset from middle to end
        if abs(self.physical_angle) > 1e-9:
            # Bent element - arc geometry (symmetric to start)
            ex = self.length * (1 - np.cos(self.physical_angle)) / (2 * self.physical_angle)
            ey = 0
            ez = self.length * np.sin(self.physical_angle) / (2 * self.physical_angle)
        else:
            # Straight element
            ex = 0
            ey = 0
            ez = self.length / 2.0

        # Local offset vector
        vec = [ex, ey, ez]

        # Rotate to global coordinates and add to middle
        end = middle + self.rotated_position(vec)
        return Position.from_list(end)


import os
from math import ceil
import pandas

magnet_table_filename = os.path.join(
    os.path.dirname(__file__), "CLARA Magnet Table v6.xlsx"
)
magnet_table = pandas.read_excel(
    magnet_table_filename,
    sheet_name="Table",
    skiprows=2,
    index_col=(2, 3, 5, 6),
    dtype={"serial number": "str"},
).fillna(0)


def create_degauss_values(maxI):
    if maxI < 10.0:
        maxI = 10.0
    return [
        round(ceil(maxI) / 10.0 * v, 2)
        for v in [9.98, -9.98, 6.0, -6.0, 4.0, -4.0, 2.0, -2.0, 1.0, -1.0, 0.0]
    ]


def add_magnet_table_parameters(n, e, magnetPV):
    try:
        magPVsplit = magnetPV.split("-")
        magnet = (
            magPVsplit[0].replace("EBT", "CLA"),
            magPVsplit[1].replace("HRG1", "GUN"),
            magPVsplit[3],
            int(magPVsplit[4]),
        )
        table = magnet_table.loc[
            magnet,
            [
                "slope [units/A]",
                "max current [A]",
                "f [units/A³]",
                "a [units/A²]",
                "I0 [A]",
                "d [units]",
                "magnetic length [mm]",
                "current [A]",
                "magnet type",
                "serial number",
            ],
        ]
        m, I_max, f, a, I0, d, L, I_degauss, manufacturer, serial_number = table
        e.magnetic.linear_saturation_coefficients.update_from_string(
            ",".join(list(map(str, [m, I_max, f, a, I0, d, L])))
        )
        e.degauss.values = create_degauss_values(float(I_degauss))
        e.manufacturer.manufacturer = manufacturer
        e.manufacturer.serial_number = serial_number
        e.electrical.maxI = ceil(I_degauss)
        e.electrical.minI = -1.0 * ceil(I_degauss)
    except Exception:
        print("Magnet missing from magnet table!", magnet)
    return e



# python
from nala.models.magnetic import (
    Multipole,
    Multipoles,
    FieldIntegral,
    LinearSaturationFit,
    MagneticElement,
    Dipole_Magnet,
    Quadrupole_Magnet,
    Sextupole_Magnet,
    Octupole_Magnet,
    Solenoid_Magnet,
    SolenoidFields,
    NonLinearLens_Magnet,
    Wiggler_Magnet,
)

def test_multipole_initialization():
    m = Multipole(order=1, normal=0.5, skew=0.2)
    assert m.order == 1
    assert m.normal == 0.5
    assert m.skew == 0.2

def test_multipoles_operations():
    multipoles = Multipoles(K1L={"order": 1, "normal": 0.5})
    assert multipoles.normal(1) == 0.5
    assert multipoles.K1L.order == 1

def test_field_integral_current_to_k():
    fi = FieldIntegral(coefficients=[1, 2, 3])
    k = fi.currentToK(current=10, energy=100)
    assert k > 0

def test_linear_saturation_fit_current_to_k():
    lsf = LinearSaturationFit(m=1, I_max=10, f=0.1, a=0.2, I0=5, d=0.3, L=1)
    result = lsf.currentToK(current=5, momentum=100)
    assert "K" in result
    assert result["K"] > 0

def test_magnetic_element_properties():
    me = MagneticElement(order=1, length=2.0, multipoles={"K1L": {"order": 1, "normal": 0.5}})
    assert me.KnL(1) == 0.5
    assert me.gradient(momentum=100) > 0

def test_dipole_magnet_properties():
    dipole = Dipole_Magnet(length=2.0)
    dipole.angle = 0.1
    assert dipole.rho > 0
    assert dipole.field_strength(momentum=100) > 0

def test_quadrupole_magnet_properties():
    quad = Quadrupole_Magnet(length=1.0, k1l=0.2)
    assert quad.k1l == 0.2

def test_sextupole_magnet_properties():
    sext = Sextupole_Magnet(length=1.0, k2l=0.2)
    assert sext.k2l == 0.2

def test_octupole_magnet_properties():
    sext = Octupole_Magnet(length=1.0, k3l=0.2)
    assert sext.k3l == 0.2

def test_solenoid_magnet_properties():
    solenoid = Solenoid_Magnet(length=1.0, ks=0.3)
    assert solenoid.ks == 0.3

def test_nonlinear_lens_magnet():
    lens = NonLinearLens_Magnet(length=1.0, integrated_strength=0.5)
    assert lens.integrated_strength == 0.5

def test_wiggler_magnet_properties():
    wiggler = Wiggler_Magnet(length=2.0, strength=1.0, period=0.1, num_periods=10)
    assert wiggler.normalized_strength > 0
    assert wiggler.poles == 20

def test_solenoid_fields():
    fields = SolenoidFields(S0L=0.5, S1L=0.3)
    assert fields.S0L == 0.5
    assert fields.S1L == 0.3


import numpy as np
import re
from warnings import warn
from .FieldParameter import FieldParameter
from ..units import UnitValue

d = ",!?/&-:;@'\n \t"


def generate_astra_field_data(self) -> np.ndarray:
    """
    Generate the field data in a format that is suitable for ASTRA, based on the
    :class:`~nala.translator.utils.fields.field` object provided.
    The `field_type` parameter determines the format of the file.
    See the `ASTRA manual`_ for more details.

    A warning is raised if the field type is not supported (perhaps elevate to a `NotImplementedError`?

    .. _ASTRA manual: https://www.desy.de/~mpyflo/Astra_manual/Astra-Manual_V3.2.pdf

    Parameters
    ----------
    self: :class:`~nala.translator.utils.fields.field`
        The field object

    Returns
    -------
    np.ndarray:
        The formatted field data.
    """
    length = str(self.length)
    data = None
    zdata = self.z_values
    if self.field_type == "LongitudinalWake":
        wzdata = self.Wz.value.val
        preamble = np.array(
            [
                [1, 0],
                [length, 0],
                [0, 0],
                [0, 0],
            ]
        )
        data = np.concatenate([preamble, np.transpose([zdata, wzdata])])
    elif self.field_type == "TransverseWake":
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        data = np.concatenate(
            [np.array([[length, ""]]), np.transpose([zdata, wxdata, wydata])]
        )
    elif self.field_type == "3DWake":
        zpreamble = np.array(
            [
                [3, 0],
                [length, 0],
                [0, 0],
                [0, 0],
            ]
        )
        xpreamble = np.array(
            [
                [length, 0],
                [0, 0],
                [0, 13],
            ]
        )
        ypreamble = np.array(
            [
                [length, 0],
                [0, 0],
                [0, 24],
            ]
        )
        wxdata = self.Wx.value.val
        wydata = self.Wy.value.val
        wzdata = self.Wz.value.val
        zvals = np.concatenate([zpreamble, np.transpose([zdata, wzdata])])
        xvals = np.concatenate([xpreamble, np.transpose([zdata, wxdata])])
        yvals = np.concatenate([ypreamble, np.transpose([zdata, wydata])])
        data = np.concatenate([zvals, xvals, yvals])
    elif self.field_type == "1DMagnetoStatic":
        bzdata = self.Bz.value.val
        data = np.transpose([zdata, bzdata])
    elif self.field_type == "3DMagnetoStatic":
        xdata = self.x.value.val
        ydata = self.y.value.val
        bzdata = self.Bz.value.val
        data = np.array(
            [
                [z, Bz]
                for x, y, z, Bz in zip(xdata, ydata, zdata, bzdata)
                if x == 0.0 and y == 0.0
            ]
        )
    elif self.field_type == "1DElectroDynamic":
        ezdata = self.Ez.value.val
        if self.cavity_type == "TravellingWave":
            spdata = ["" for _ in range(self.length)]
            if any(
                    [
                        getattr(self, param) is None for param in [
                        "start_cell_z",
                        "end_cell_z",
                        "mode_numerator",
                        "mode_denominator"
                    ]
                    ]
            ):
                raise ValueError(
                    "start_cell_z, end_cell_z", "mode_numerator", "mode_denominator"
                                                                  "must be defined for TravellingWave cavities"
                )
            preamble = np.array(
                [
                    [
                        self.start_cell_z,
                        self.end_cell_z,
                        self.mode_numerator,
                        self.mode_denominator,
                    ]
                ]
            )
            data = np.concatenate(
                [preamble, np.transpose([zdata, ezdata, spdata, spdata])]
            )
        else:
            data = np.transpose([zdata, ezdata])
    elif self.field_type == "2DElectroDynamic":
        warn(f"Converting 2DElectroDynamic field map for {self.filename} to 1D for ASTRA")
        idx = np.where(self.r.value.val == 0)
        zdata = self.z.value.val[idx]
        ezdata = self.Ez.value.val[idx]
        data = np.transpose([zdata, ezdata])
    elif self.field_type == "1DQuadrupole":
        gdata = self.G.value.val
        data = np.transpose([zdata, gdata])
    else:
        warn(f"Field type {self.field_type} not supported for ASTRA")
    return data


def write_astra_field_file(self) -> str:
    """
    Write the field data in an ASTRA-compatible format, based on the
    :class:`~SimulationFramework.Modules.Fields.field` object provided.
    The absolute location of the file to be written is generated using
    :func:`~SimulationFramework.Modules.Fields.field._output_filename`, which is parsed from the Master Lattice.

    Parameters
    ----------
    self: :class:`~SimulationFramework.Modules.Fields.field`
        The field object

    Returns
    -------
    str:
        The converted filename
    """
    astra_file = self._output_filename(extension=".astra")
    data = generate_astra_field_data(self)
    if data is not None:
        with open(f"{astra_file}", "w") as f:
            for d in data:
                f.write(" ".join([str(x) for x in d]) + "\n")
    return astra_file


def read_astra_field_file(
    self,
    filename: str,
    field_type: str,
    cavity_type: str | None = None,
    frequency: float | None = None,
):
    """
    Read a field file from ASTRA format and convert it into a
    :class:`~SimulationFramework.Modules.Fields.field` object (self).
    Certain parameters must be included, particularly for RF cavities.

    See the `ASTRA manual`_ for more details.

    .. _ASTRA manual: https://www.desy.de/~mpyflo/Astra_manual/Astra-Manual_V3.2.pdf

    Parameters
    ----------
    self: :class:`~SimulationFramework.Modules.Fields.field`
        The field object to be updated.
    filename: str
        The path to the ASTRA field file
    field_type: str
        The name of the field, see :attr:`~SimulationFramework.Modules.Fields.allowed_fields`
    cavity_type: str, optional
        The type of RF cavity, see :attr:`~SimulationFramework.Modules.Fields.allowed_cavities`
    frequency: float, optional
        The frequency of the RF cavity.

    Returns
    -------
    None

    Raises
    ------
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `cavity_type` is not provided
    ValueError:
        if the cavity `field_type` contains the string `Electro` and `frequency` is not provided
    NotImplementedError:
        if a given `field_type` is not implemented
    """
    self.reset_dicts()
    setattr(self, "field_type", field_type)
    try:
        if "Electro" in field_type:
            if cavity_type is None:
                raise ValueError(f"cavity_type must be provided for {field_type}")
            else:
                setattr(self, "cavity_type", cavity_type)
            if frequency is None:
                raise ValueError(f"frequency must be provided for {field_type}")
            else:
                setattr(self, "frequency", frequency)
    except Exception:
        raise ValueError(
            f"Fields read_astra_field_file error: {filename}, {field_type}, {cavity_type}, {frequency}"
        )
    if field_type == "1DMagnetoStatic":
        fdat = np.loadtxt(filename)
        setattr(
            self, "z", FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m"))
        )
        setattr(
            self,
            "Bz",
            FieldParameter(
                name="Bz", value=UnitValue(fdat[::, 1] / np.max(fdat[::, 1]), units="T")
            ),
        )
    elif field_type == "1DElectroDynamic":
        if cavity_type == "StandingWave":
            fdat = np.loadtxt(filename)
            setattr(
                self,
                "z",
                FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m")),
            )
            setattr(
                self,
                "Ez",
                FieldParameter(
                    name="Ez",
                    value=UnitValue(fdat[::, 1] / np.max(fdat[::, 1]), units="V/m"),
                ),
            )
        elif cavity_type == "TravellingWave":
            with open(filename) as f:
                rl = f.readlines()[0]
                twdat = re.split("[" + "\\".join(d) + "]", rl)
                setattr(self, "start_cell_z", float(twdat[0]))
                setattr(self, "end_cell_z", float(twdat[1]))
                setattr(self, "mode_numerator", float(twdat[2]))
                setattr(self, "mode_denominator", float(twdat[3]))
            fdat = np.loadtxt(filename, skiprows=1)
            setattr(
                self,
                "z",
                FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m")),
            )
            setattr(
                self,
                "Ez",
                FieldParameter(
                    name="Ez",
                    value=UnitValue(fdat[::, 1] / np.max(fdat[::, 1]), units="V/m"),
                ),
            )
    elif field_type == "LongitudinalWake":
        try:
            fdat = np.loadtxt(filename)
            numrows = int(fdat[1, 0])
            setattr(
                self,
                "z",
                FieldParameter(
                    name="z", value=UnitValue(fdat[4 : 4 + numrows][::, 0], units="m")
                ),
            )
            setattr(
                self,
                "Wz",
                FieldParameter(
                    name="Wz",
                    value=UnitValue(fdat[4 : 4 + numrows][::, 1]),
                    units="V/C",
                ),
            )
        except Exception:
            fdat = np.loadtxt(filename, skiprows=1)
            setattr(
                self,
                "z",
                FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m")),
            )
            setattr(
                self,
                "Wz",
                FieldParameter(name="Wz", value=UnitValue(fdat[::, 1]), units="V/C"),
            )
    elif field_type == "3DWake":
        fdat = np.loadtxt(filename)
        numrows = int(fdat[1, 0])
        setattr(
            self,
            "z",
            FieldParameter(
                name="z", value=UnitValue(fdat[4 : 4 + numrows][::, 0], units="m")
            ),
        )
        setattr(
            self,
            "Wz",
            FieldParameter(
                name="Wz", value=UnitValue(fdat[4 : 4 + numrows][::, 1]), units="V/C"
            ),
        )
        setattr(
            self,
            "Wx",
            FieldParameter(
                name="Wx",
                value=UnitValue(fdat[numrows + 7 : (2 * numrows) + 7][::, 1]),
                units="V/C/m",
            ),
        )
        setattr(
            self,
            "Wy",
            FieldParameter(
                name="Wy",
                value=UnitValue(fdat[(2 * numrows) + 10 :][::, 1]),
                units="V/C/m",
            ),
        )
    elif field_type == "1DQuadrupole":
        fdat = np.loadtxt(filename)
        setattr(
            self, "z", FieldParameter(name="z", value=UnitValue(fdat[::, 0], units="m"))
        )
        setattr(
            self,
            "G",
            FieldParameter(
                name="G",
                value=UnitValue(fdat[::, 1] / np.max(fdat[::, 1]), units="T/m"),
            ),
        )
    else:
        raise NotImplementedError(
            f"{field_type} loading not implemented for ASTRA files"
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#
# Project information
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
#
# General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
#
# Options for HTML output
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
#
# Options for Pydantic models
# https://autodoc-pydantic.readthedocs.io/en/stable/users/configuration.html

import os
import sys
import subprocess

# docs/ is at project_root/docs
# add project root:
sys.path.insert(0, os.path.abspath('..'))
# also add project_root/nala so Python can find the inner package dir
sys.path.insert(0, os.path.abspath('../nala'))



def get_git_revision_hash(short: bool = True) -> str:
    """Returns the hash representing the current git revision

    :param bool short: Return the short version of the hash
    :returns: Hash representing the current local commit
    """
    command = "git rev-parse --short HEAD" if short else "git rev-parse HEAD"
    revision = subprocess.check_output(command.split())
    return revision.decode("ascii").strip()


def get_git_revision_tags() -> str:
    """Returns tags associated with the current git revision

    :returns: Tags assigned the the current local commit
    """
    command = "git tag --points-at HEAD"
    tags = subprocess.check_output(command.split())
    if len(tags) != 0:
        return tags.decode("ascii").strip()
    else:
        return None


def skip_member(app, what, name, obj, skip, options):
    if name == "Field":  # or any attribute you find problematic
        return True
    return skip

# Prevent sphinx-autodoc-typehints from trying to process pydantic.Field
def suppress_pydantic_field_signature(app, what, name, obj, options, signature, return_annotation):
    # Returning (signature, return_annotation) unchanged disables modification
    return (None, None)

def setup(app):
    # Keep your skip_member hook
    app.connect("autodoc-skip-member", skip_member)
    # Add this to stop autodoc-typehints signature unwrapping
    app.connect("autodoc-process-signature", suppress_pydantic_field_signature)



# # use custom theme to set html width
# def setup(app):
#     app.add_css_file("my_theme.css")


# set relative path to the documented project
sys.path.insert(0, os.path.abspath("../../"))

# set project name, author and copyright
author = "James Jones, Alex Brynes, Matthew King, Mark Johnson"
copyright = "2025, STFC ASTeC"
project = "NALA"

# fetch version (commit hash or release tags) from git
_git_tags = get_git_revision_tags()
version = get_git_revision_hash(short=True) if _git_tags is None else _git_tags

# set sphinx extensions
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    "sphinx_autodoc_typehints",
    #"sphinxcontrib.autodoc_pydantic",
    "sphinx.ext.mathjax",  # enables math rendering
    #"myst_nb",
]


# set sphinx options
source_suffix = ".rst"  # use reStructedText files for sphinx pages
master_doc = "index"  # name for the root document
exclude_patterns = ["_build"]  # patterns to exclude when looking for source files
templates_path = ["_templates"]  # list of paths that contain extra templates
add_function_parentheses = True  # display function and method names with parentheses
add_module_names = False  # don't include module names before object names
pygments_style = "sphinx"  # style for highlighting of source code
bibtex_bibfiles = ["references.bib"]

# set automodapi options
automodapi_toctreedirnm = (
    "autodoc"  # relative path to automodapi generated documentation
)
automodapi_writereprocessed = False  # special flag for automodapi debugging
typehints_fully_qualified = False  # don't include module names before object names
always_document_param_types = True
typehints_document_rtype = True  # adds :rtype: directives for autodoc type hints

automodsumm_inherited_members = (
    False  # don't include inherited members in documentations for classes
)

# set options for pydantic models
autodoc_pydantic_model_show_json = (
    False  # don't include JSON schema for pydantic models
)
autodoc_pydantic_model_show_field_summary = (
    True  # don't include a bullet-point list of model fields
)
autodoc_pydantic_model_show_config_summary = (
    False  # don't include model configurations for pydantic models
)
autodoc_pydantic_field_list_validators = (
    True  # don't list validators for pydantic model fields
)
autodoc_pydantic_field_show_constraints = (
    False  # don't list constraints for pydantic model fields
)
autodoc_pydantic_model_show_validator_summary = (
    False  # dont' include validator methods for pydantic models
)
autodoc_pydantic_model_show_validator_members = (
    False  # don't include documentation for validator methods
)
autodoc_pydantic_field_doc_policy = "both"  # shows docstrings and Field descriptions
#autodoc_typehints = "none"

autodoc_mock_imports = [
    #"nala.models.elementList",  # or whichever module fails
    "pydantic",                 # mock dependencies if needed
]

numfig = True


autodoc_default_options = {
    "exclude-members": ",".join([
        "model_post_init",
        "model_fields",
        "model_config",
        "model_computed_fields",
        "model_dump",
        "model_dump_json",
        "model_validate",
        "model_validate_json",
        "model_copy",
        "model_json_schema",
        "construct",
        "schema",
        "schema_json",
        "dict",
        "json",
    ]),
    "members": True,
    "undoc-members": True,
    "inherited-members": False,
}


autodoc_pydantic_settings_show_config_summary = (
    False
)

autodoc_pydantic_field_list_validators = False
autodoc_pydantic_model_show_validator_members = False

#nb_execution_mode = "off"  # options: "off", "auto", "force"


# mapping to other projects' documentation pages
intersphinx_mapping = {
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "python": ("https://docs.python.org/3", None),
}

# set HTML output options
html_theme = "sphinx_rtd_theme"  # set HTML themse (read the docs theme)
html_logo = "icon.png"  # set logo for top-left corner of HTML pages
html_favicon = "favicon.ico"  # set HTML favicon
html_static_path = ["_static"]  # path to custom static files

# set mathjax equation rendering options
mathjax3_config = {"chtml": {"displayAlign": "left", "displayIndent": "2em"}}

# set date and time format string
today_fmt = "%A %B %d %Y (%H:%M:%S)"



import os
import numpy as np
import yaml
from pydantic import BaseModel, ConfigDict
from typing import Dict
from scipy.spatial.transform import Rotation
from ocelot.cpbd.magnetic_lattice import MagneticLattice
import nala.models.element as NALA_elements
from . import magnetic_orders
from ...utils.functions import introspect_model_defaults
from ...conversion_rules.codes import ocelot_conversion
from warnings import warn

type_conversion_rules_Ocelot = ocelot_conversion.ocelot_conversion_rules


with open(
    os.path.dirname(os.path.abspath(__file__)) +
    "/../../conversion_rules/keywords/keyword_conversion_rules_ocelot.yaml",
    "r",
) as infile:
    keyword_conversion_rules = yaml.safe_load(infile)

class OcelotLatticeImporter(BaseModel):

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
    )

    name: str = "Lattice"

    machine_area: str = "Lattice"

    magnetic_lattice: MagneticLattice
    """Name of ELEGANT parameters file"""

    nala_elements: Dict = {}
    """Dictionary containing converted element objects"""

    def magnetic_lattice_to_elements(self):
        return self.lattice_to_cartesian_with_rotation(self.magnetic_lattice.sequence)

    def create_element_dictionary(self):
        elements = self.magnetic_lattice_to_elements()
        self.nala_elements = {}
        switch_dict = {f"{str(y).lower().split('.')[-1].strip("\'>")}_{x}": x for x, y in
                       type_conversion_rules_Ocelot.items()}

        for elem, pos_and_rot in elements.items():
            # if type(elem) not in switch_dict:
            #     warn(f"Ocelot element type {type(elem)} not convertible")
            typeconv = str(type(elem)).lower().split('.')[-1].strip("\'>")
            key = None
            for k, v in switch_dict.items():
                if v == k.split('_')[-1] and typeconv in k:
                    key = k
            if not key:
                warn(f"Could not find element type {type(elem)} for {elem.id}; "
                     f"setting as drift")
                key = "drift_drift"
            newobj = {
                "name": elem.id,
                "hardware_type": switch_dict[key],
                "hardware_class": switch_dict[key],
                "machine_area": self.machine_area}
            try:
                merged = keyword_conversion_rules[switch_dict[key].lower()] | keyword_conversion_rules["general"]
            except KeyError:
                merged = keyword_conversion_rules["general"]
            for sfparam, oceparam in merged.items():
                if hasattr(elem, oceparam):
                    newobj.update({sfparam: getattr(elem, oceparam)})
            sftype = switch_dict[key]
            try:
                if sftype == "Kicker":
                    model_fields = introspect_model_defaults(getattr(NALA_elements, "Combined_Corrector"))
                    newobj["hardware_type"] = "Combined_Corrector"
                elif "Cavity" not in sftype:
                    model_fields = introspect_model_defaults(getattr(NALA_elements, sftype.capitalize()))
                    newobj["hardware_type"] = sftype.capitalize()
                else:
                    model_fields = introspect_model_defaults(getattr(NALA_elements, sftype))
            except AttributeError:
                print(f"type {sftype} not recognized")
                newobj.update(
                    {
                        k: {
                            "hardware_type": "Drift",
                            "name": k,
                            "hardware_class": "Drift",
                            "machine_area": self.machine_area
                        }
                    }
                )
                continue
            for subk in ["magnetic", "cavity", "simulation", "diagnostic", "physical"]:
                if subk in model_fields:
                    newobj.update({subk: {}})
            for oceparam, value in elem.element.__dict__.items():
                oceparam = oceparam.lower()
                kwele = {y: x for x, y in merged.items()}
                for subk in model_fields:
                    if isinstance(model_fields[subk], dict):
                        if oceparam in ["k1", "k2", "k3", "angle"] and newobj['hardware_type'] in magnetic_orders:
                            if "magnetic" not in newobj:
                                newobj.update({"magnetic": {}})
                            try:
                                newobj["magnetic"]["kl"] = getattr(
                                    elem.element, f"k{magnetic_orders[newobj['hardware_type']]}"
                                ) * elem.l
                            except AttributeError:
                                newobj["magnetic"]["kl"] = elem.element.angle
                            newobj["magnetic"].update({oceparam: value})
                            newobj["hardware_class"] = "Magnet"
                        if oceparam in model_fields[subk] and hasattr(elem, oceparam):
                            newobj[subk].update({oceparam: getattr(elem, oceparam)})
                        elif oceparam in kwele:
                            if kwele[oceparam] in model_fields[subk]:
                                if not isinstance(model_fields[subk][kwele[oceparam]], str) or model_fields[subk][
                                    kwele[oceparam]]:
                                    try:
                                        if oceparam == "v" and "Cavity" in newobj["hardware_type"]:
                                            newobj["hardware_class"] = "RF"
                                            newobj[subk].update({kwele[oceparam]: getattr(elem, oceparam) * 1e9})
                                        else:
                                            newobj[subk].update({kwele[oceparam]: getattr(elem, oceparam)})
                                    except KeyError:
                                        pass
                                    except AttributeError:
                                        pass
            pos = pos_and_rot[0][::-1]
            rot = [float(pos_and_rot[1][0]), float(pos_and_rot[1][2]), float(pos_and_rot[1][1])]
            newobj["physical"]["position"] = pos
            newobj["physical"]["global_rotation"] = rot
            self.nala_elements.update({elem.id: getattr(NALA_elements, newobj["hardware_type"])(**newobj)})

    def save_lattice_file(self, filename: str, directory: str):
        if not self.nala_elements:
            self.create_framework_element_dictionary()
        save_lattice_file(self.nala_elements, filename, directory)

    @staticmethod
    def lattice_to_cartesian_with_rotation(elements) -> Dict:
        """
        Compute Cartesian coordinates [x, y, z] of accelerator lattice elements
        and the global rotation (Euler angles) at the MIDPOINT of each element.
        """

        x, y, z = 0.0, 0.0, 0.0
        theta_h = 0.0
        theta_v = 0.0
        elems, positions, rotations = [], [], []
        cumulative_R = np.eye(3)

        for elem in elements:
            if "bend" not in str(type(elem)).lower() or abs(getattr(elem, "angle", 0.0)) < 1e-9:
                # --- Drift ---
                L = elem.l
                # Direction vector
                dx = L * np.cos(theta_v) * np.cos(theta_h)
                dy = L * np.sin(theta_v)
                dz = L * np.cos(theta_v) * np.sin(theta_h)

                # Midpoint is halfway along the segment
                mid_x = x + dx / 2
                mid_y = y + dy / 2
                mid_z = z + dz / 2

                # Store midpoint
                euler_angles = Rotation.from_matrix(cumulative_R).as_euler('zyx', degrees=False)
                elems.append(elem)
                positions.append(np.array([mid_x, mid_y, mid_z]))
                rotations.append(euler_angles)

                # Move to exit for next element
                x += dx
                y += dy
                z += dz

            else:
                # --- Dipole Bend ---
                L, phi, tilt = elem.l, elem.angle, elem.tilt

                if np.isclose(tilt, 0):  # Horizontal bend (x-z plane)
                    R_bend = Rotation.from_euler('y', phi).as_matrix()
                    R_half = Rotation.from_euler('y', phi / 2).as_matrix()

                    R_geom = L / phi  # bending radius

                    # Center of curvature
                    cx = x - R_geom * np.sin(theta_h)
                    cz = z + R_geom * np.cos(theta_h)

                    # Midpoint (half of bend angle)
                    theta_mid = theta_h + phi / 2
                    mid_x = cx + R_geom * np.sin(theta_mid)
                    mid_y = y
                    mid_z = cz - R_geom * np.cos(theta_mid)

                    # Rotation halfway through bend
                    R_mid = cumulative_R @ R_half

                    euler_angles = Rotation.from_matrix(R_mid).as_euler('zyx', degrees=False)
                    elems.append(elem)
                    positions.append(np.array([mid_x, mid_y, mid_z]))
                    rotations.append(euler_angles)

                    # Update to exit of element
                    theta_h += phi
                    x = cx + R_geom * np.sin(theta_h)
                    z = cz - R_geom * np.cos(theta_h)
                    cumulative_R = cumulative_R @ R_bend

                elif np.isclose(tilt, np.pi / 2):  # Vertical bend (x-y plane)
                    R_bend = Rotation.from_euler('x', -phi).as_matrix()
                    R_half = Rotation.from_euler('x', -phi / 2).as_matrix()
                    R_geom = L / phi

                    cy = y - R_geom * np.cos(theta_v)

                    # Midpoint
                    theta_mid = theta_v + phi / 2
                    mid_y = cy + R_geom * np.cos(theta_mid)
                    mid_x = x
                    mid_z = z
                    R_mid = cumulative_R @ R_half

                    euler_angles = Rotation.from_matrix(R_mid).as_euler('zyx', degrees=False)
                    elems.append(elem)
                    positions.append(np.array([mid_x, mid_y, mid_z]))
                    rotations.append(euler_angles)

                    # Exit of element
                    theta_v += phi
                    y = cy + R_geom * np.cos(theta_v)
                    cumulative_R = cumulative_R @ R_bend

                else:
                    raise ValueError(f"Unrecognized tilt angle {tilt} for {elem.id}")

        positions_and_rotations = [(p, r) for p, r in zip(np.array(positions), np.array(rotations))]
        return {e: pr for e, pr in zip(elems, positions_and_rotations)}

    @staticmethod
    def _convert_k_to_kl(v) -> dict:
        newk = {}
        for n in range(1, 9):
            if hasattr(v, f"k{n}") and hasattr(v, "l"):
                newk.update({f"k{n}l": getattr(v, f"k{n}") * getattr(v, "l")})
        return newk


from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.parser import Parser
from yaml.composer import Composer
from Importers.MySafeConstructor import MySafeConstructor
from yaml.resolver import Resolver


class MySafeLoader(Reader, Scanner, Parser, Composer, MySafeConstructor, Resolver):

    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        MySafeConstructor.__init__(self)
        Resolver.__init__(self)



from typing import Dict, Any
from warnings import warn
import numpy as np
from pydantic import PositiveInt

from ...models.elementList import SectionLattice
from ...models.RF import WakefieldElement
from ...models.simulation import WakefieldSimulationElement, DiagnosticSimulationElement
from .cavity import RFCavityTranslator
from .converter import translate_elements
from .diagnostic import DiagnosticTranslator
from .wake import WakefieldTranslator
from .codes.gpt import gpt_ccs, gpt_Zminmax, gpt_dtmint
from ..utils.functions import tw_cavity_energy_gain
from ..utils.fields import field


class SectionLatticeTranslator(SectionLattice):
    """
    Translator class for converting a :class:`~nala.models.elementList.SectionLattice` instance into a string or
    object that can be understood by various simulation codes.
    """

    directory: str = '.'
    """Directory to which files will be written."""

    astra_headers: Dict = {}
    """Headers for ASTRA input file; see :class:`~nala.translator.converters.codes.astra.astra_header`
    and its child classes."""

    csrtrack_headers: Dict = {}
    """Headers for CSRTrack input file; see :class:`~nala.translator.converters.codes.csrtrack.csrtrack_element`
    and its child classes.."""

    gpt_headers: Dict = {}
    """Headers for GPT input file; see :class:`~nala.translator.converters.codes.astra.gpt_element`
    and its child classes.."""

    opal_headers: Dict = {}
    """Headers for OPAL input file; see :class:`~nala.translator.converters.codes.opal.opal_header`
    and its child classes..
    
    WARNING: OPAL not fully benchmarked / tested.
    """

    csr_enable: bool = True
    """Flag to enable calculation of CSR in drifts."""

    lsc_enable: bool = True
    """Flag to enable calculation of LSC in drifts."""

    lsc_bins: PositiveInt = 20
    """Number of LSC bins for drifts."""

    @classmethod
    def from_section(cls, section: SectionLattice) -> "SectionLatticeTranslator":
        """
        Method for creating an instance of this class based on an existing
        :class:`~nala.models.elementList.SectionLattice`.

        Parameters
        ----------
        section: SectionLattice
            The existing :class:`~nala.models.elementList.SectionLattice`

        Returns
        -------
        :class:`~nala.translator.converters.section.SectionLatticeTranslator`
            An instance of this class.

        """
        return cls.model_validate(
            {
                "name": section.model_copy().name,
                "order": section.model_copy().order,
                "elements": section.model_copy().elements,
                "master_lattice_location": section.model_copy().master_lattice_location
            }
        )

    def to_astra(self) -> str:
        """
        Create an ASTRA-compatible input file based on the lattice information and
        the settings provided in :attr:`~astra_headers`.

        Returns
        -------
        str
            An ASTRA-compatible input file.
        """
        from .codes.astra import section_header_text_ASTRA
        headers = ["&APERTURE", "&CAVITY", "&SOLENOID", "&QUADRUPOLE", "&DIPOLE", "&WAKE"]
        counter = {k: 1 for k in headers}
        written = []
        element_headers = {h: "" for h in headers}
        elem_dict = translate_elements(
            list(self.elements.elements.values()),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        astrastr = ""
        for h in self.astra_headers.values():
            astrastr += h.write_ASTRA()

        for e in elem_dict.values():
            for key, count in counter.items():
                if "&" + e.hardware_type.upper().replace("RF", "").replace("FIELD", "") == key:
                    if key not in written:
                        element_headers[key] += f"{section_header_text_ASTRA[key]} = True\n"
                        written.append(key)
                    element_headers[key] += e.to_astra(n=count)
                    counter[key] += 1
                    try:
                        w = WakefieldTranslator(
                            name=e.name + "_wake",
                            hardware_class="Wakefield",
                            hardware_type="Wakefield",
                            machine_area=e.machine_area,
                            physical=e.physical,
                            cavity=WakefieldElement(cell_length=e.cavity.cell_length, n_cells=e.cavity.n_cells),
                            simulation=WakefieldSimulationElement(
                                wakefield_definition=e.simulation.wakefield_definition),
                            directory=e.directory,
                        )
                        if "&WAKE" not in written:
                            element_headers["&WAKE"] += f"{section_header_text_ASTRA["&WAKE"]} = True\n"
                            written.append("&WAKE")
                        element_headers["&WAKE"] += w.to_astra(n=counter["&WAKE"])
                        counter["&WAKE"] += e.cavity.n_cells
                    except Exception as ex:
                        pass
                else:
                    cond = "&" + e.hardware_type.upper().replace("RF", "").replace("FIELD", "") in headers
                    if not e.hardware_class == "Diagnostic" and not cond:
                        warn(f"Element of type {e.hardware_type} not supported for ASTRA")
        for k, v in element_headers.items():
            astrastr += k + "\n"
            astrastr += v + "\n"
            astrastr += "/ \n"
        return astrastr

    def to_gpt(self, startz: float, endz: float, Brho: float = 0.0, dtmin: float | None = None) -> str:
        """
        Create a GPT-compatible input file based on the lattice information and
        the settings provided in :attr:`~gpt_headers`.

        Note that, for sections with accelerating sections, the magnetic rigidity `Brho` may not
        be updated correctly, which may affect the accuracy of tracking through dipoles.

        Parameters
        ----------
        startz: float
            Start longitudinal location of the lattice.
        endz: float
            End longitudinal location of the lattice.
        Brho: float
            Magnetic rigidity.
        dtmin: float, optional
            Minimum time step size for integration

        Returns
        -------
        str
            A GPT-compatible input file.
        """
        fulltext = ""
        for header in self.gpt_headers.values():
            fulltext += header.write_GPT()
        elem_dict = translate_elements(
            list(self.elements.elements.values()),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        for i, element in enumerate(list(elem_dict.values())):
            if i == 0:
                ccs = gpt_ccs(
                    name="wcs",
                    position=element.physical.start.model_dump(),
                    rotation=element.physical.global_rotation.model_dump(),
                )
            fulltext += element.to_gpt(Brho, ccs=ccs.name)
            if (
                element.hardware_type.lower() == "rfcavity"
                and isinstance(element.simulation.wakefield_definition, field)
            ):
                w = WakefieldTranslator(
                    name=element.name + "_wake",
                    hardware_class="Wakefield",
                    hardware_type="Wakefield",
                    machine_area=element.machine_area,
                    physical=element.physical,
                    cavity=WakefieldElement(cell_length=element.cavity.cell_length, n_cells=element.cavity.n_cells),
                    simulation=WakefieldSimulationElement(
                        wakefield_definition=element.simulation.wakefield_definition,
                    ),
                    directory=element.directory,
                )
                fulltext += w.to_gpt(Brho, ccs=ccs.name)
            new_ccs = element.ccs
            if not new_ccs.name == ccs.name:
                relpos, relrot = ccs.relative_position(
                    element.physical.middle.model_dump(), element.physical.global_rotation.model_dump()
                )
            else:
                relpos = element.physical.middle.model_dump()
            screen0pos = 0
            ccs = new_ccs
            if element.hardware_class.lower() == "diagnostic":
                fulltext += f'screen({ccs.name_as_str}, "I", {str(relpos[2])}, {ccs.name_as_str});\n'
                # if self.gpt_headers["setfile"].particle_definition == "laser":
        lastelem = list(elem_dict.values())[-1]
        lastscreen = DiagnosticTranslator(
            name="end_screen",
            hardware_class="Diagnostic",
            hardware_type="Diagnostic",
            machine_area=lastelem.machine_area,
            simulation=DiagnosticSimulationElement(output_filename=f"{self.name}_out.gdf"),
            physical=lastelem.physical,
        )
        fulltext += lastscreen.to_gpt(
            Brho, ccs=ccs.name, output_ccs="wcs"
        )
        relpos, relrot = ccs.relative_position(
            lastelem.physical.end.model_dump(), lastelem.physical.global_rotation.model_dump()
        )
        fulltext += f'screen({ccs.name_as_str}, "I", {str(relpos[2])}, {ccs.name_as_str});\n'
        zminmax = gpt_Zminmax(
            ECS='"wcs", "I"',
            zmin=startz - 0.1,
            zmax=endz + 1,
        )
        fulltext += zminmax.write_GPT()
        if dtmin is not None:
            dtmint = gpt_dtmint(dtmin=dtmin)
            fulltext += dtmint.write_GPT()
        return fulltext

    def to_opal(self, energy: float = 0, breakstr: str="") -> str:
        """
        Create an OPAL-compatible input file based on the lattice information and
        the settings provided in :attr:`~opal_headers`.

        Note that, for sections with accelerating sections, the beam energy `energy` may not
        be updated correctly, which may affect the accuracy of tracking through dipoles.

        Parameters
        ----------
        energy: float
            Beam energy
        breakstr: str
            String for separating sections in the lattice file.

        Returns
        -------
        str
            An OPAL-compatible input file.
        """
        check_dict = [
            "option",
            "distribution",
            "fieldsolver",
            "beam",
            "track",
            "run",
        ]
        for k in check_dict:
            if k not in self.opal_headers:
                raise KeyError(f"Header {k} must be defined for OPAL.")
        fulltext = ""
        fulltext += self.opal_headers["option"].write_Opal()
        fulltext += f'{breakstr}\n// LATTICE\n'
        zstops = []
        elem_dict = translate_elements(
            list(self.elements.elements.values()),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        written = []
        svals = self.get_s_values(as_dict=True, at_entrance=True)
        for d in elem_dict.values():
            if isinstance(d, RFCavityTranslator):
                if d.structure_type.lower() == "travellingwave":
                    energy += tw_cavity_energy_gain(d)
                else:
                    energy += d.field_amplitude * np.cos(np.pi * d.phase / 180)
            sval = d.physical.start.z if d.subelement else svals[d.name]
            stnew = d.to_opal(sval=sval, designenergy=energy)
            if len(stnew) > 0:
                written.append(d.name)
                fulltext += d.to_opal(sval=sval, designenergy=energy)
            zstops.append(d.physical.end.z)
        zstop = max(zstops)
        self.opal_headers["track"].ZSTOP = zstop
        fulltext += "\n" + self.name + ": LINE=("
        for e, element in list(elem_dict.items()):
            if len((fulltext + e).splitlines()[-1]) > 60:
                fulltext += "\n"
            if element.name in written:
                fulltext += e.replace('-', '_') + ", "
        fulltext = (fulltext[:-2] + ");\n")

        fulltext += self.opal_headers["distribution"].write_Opal()
        fulltext += self.opal_headers["fieldsolver"].write_Opal()
        fulltext += self.opal_headers["beam"].write_Opal()
        fulltext += self.opal_headers["track"].write_Opal()
        fulltext += self.opal_headers["run"].write_Opal()
        fulltext += "ENDTRACK;\n\n Quit;\n"
        return fulltext

    def to_elegant(self, charge: float = None) -> str:
        """
        Create an ELEGANT-compatible input file based on the lattice information.

        Parameters
        ----------
        charge: float
            Bunch charge

        Returns
        -------
        str
            An ELEGANT-compatible lattice file.
        """
        section_with_drifts = self.createDrifts(
            csr_enable=self.csr_enable,
            lsc_enable=self.lsc_enable,
            lsc_bins=self.lsc_bins,
        )
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        string = ""
        if charge:
            string += f"{self.name}_Q: CHARGE, TOTAL = {charge};\n"

        for d in elem_dict.values():
            string += d.to_elegant()

        string += f"{self.name}: LINE = ("
        if charge:
            string += f"{self.name}_Q, "
        for elem in section_with_drifts.keys():
            string += f"{elem}, "
        string = f"{string[:-2]})" + "\n"
        return string

    def to_genesis(self) -> str:
        """
        Create a Genesis-compatible input file based on the lattice information.

        Returns
        -------
        str
            A Genesis-compatible lattice file (v4).
        """
        section_with_drifts = self.createDrifts()
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        string = ""

        for d in elem_dict.values():
            string += d.to_genesis()

        string += f"{self.name}: LINE = " + "{"
        for elem in section_with_drifts.keys():
            string += f"{elem}, "
        string = f"{string[:-2]}" + "};\n"
        return string

    def to_ocelot(self, save=False) -> "MagneticLattice":
        """
        Create an Ocelot-compatible magnetic lattice object based on the lattice information.

        Parameters
        ----------
        save: bool
            Flag to indicate whether to save the lattice to a file.

        Returns
        -------
        MagneticLattice
            An Ocelot `MagneticLattice` object.
        """
        from ocelot.cpbd.magnetic_lattice import MagneticLattice
        from ocelot.cpbd.transformations.second_order import SecondTM
        from ocelot.cpbd.transformations.kick import KickTM
        from ocelot.cpbd.transformations.runge_kutta import RungeKuttaTM
        from ocelot.cpbd.elements import Octupole, Undulator, Marker
        method = {"global": SecondTM, Octupole: KickTM, Undulator: RungeKuttaTM}
        section_with_drifts = self.createDrifts()
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        elements = []

        for d in elem_dict.values():
            elements.append(d.to_ocelot())

        maglat = MagneticLattice(elements, method=method)
        if save:
            maglat.save_as_py_file(f"{self.directory}/{self.name}.py")

        return maglat

    def to_cheetah(self, save=False) -> "Segment":
        """
        Create a Cheetah-compatible lattice segment object based on the lattice information.

        Parameters
        ----------
        save: bool
            Flag to indicate whether to save the lattice to a file.

        Returns
        -------
        Segment
            A Cheetah `Segment` object.
        """
        from cheetah import Segment
        section_with_drifts = self.createDrifts()
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        segment = []
        segments = False
        for element in elem_dict.values():
            if not element.subelement:
                elem = element.to_cheetah()
                if elem is not None:
                    segment.append(elem)
                    segments = True
        if segments:
            full_segment = Segment(elements=segment, name=self.name)
        else:
            raise ValueError(f"No cheetah elements added for {self.name}")

        if save:
            full_segment.to_lattice_json(
                filepath=f'{self.directory}/{self.name}.json'
        )

        return full_segment

    def to_xsuite(self, beam_length: int, env: Any = None, particle_ref: Any = None, save=True) -> "Line":
        """
        Create an Xsuite-compatible lattice line object based on the lattice information.

        Parameters
        ----------
        beam_length: int
            Number of particles in the beam
        env: xtrack.Environment
            xtrack Environment object; if `None`, it will be created
        particle_ref: xtrack.Particles
            xtrack Particles object
        save: bool
            Flag to indicate whether to save the `Line` to JSON.

        Returns
        -------
        Segment
            A Xsuite `Line` object.
        """
        import xtrack as xt
        if not isinstance(env, xt.Environment):
            env = xt.Environment()
        section_with_drifts = self.createDrifts()
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        line = env.new_line()
        for i, element in enumerate(list(elem_dict.values())):
            if not element.subelement:
                name, component, properties = element.to_xsuite(
                    beam_length=beam_length
                )
                line.append(element.name, component(**properties))
        if isinstance(particle_ref, xt.Particles):
            line.particle_ref = particle_ref
        if save:
            line.to_json(f"{self.directory}/{self.name}.json")
        return line

    def to_csrtrack(self) -> str:
        """
        Create a CSRTrack-compatible input file based on the lattice information and
        the settings provided in :attr:`~csrtrack_headers`.

        Returns
        -------
        str
            A CSRTrack-compatible lattice file.
        """
        headers = ["dipole", "quadrupole", "screen"]
        counter = {k: 1 for k in headers}
        elem_dict = translate_elements(
            list(self.elements.elements.values()),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        csrtrackstr = "io_path{logfile = log.txt}\nlattice{\n"
        for e in elem_dict.values():
            for key, count in counter.items():
                if e.hardware_type.lower() == key:
                    csrtrackstr += e.to_csrtrack(n=count)
                    counter[key] += 1
                else:
                    if not e.hardware_class == "Diagnostic":
                        warn(f"Element of type {e.hardware_type} not supported for CSRTrack")
        lastelem = list(elem_dict.values())[-1]
        lastscreen = DiagnosticTranslator(
            name="end_screen",
            hardware_class="Diagnostic",
            hardware_type="Diagnostic",
            machine_area=lastelem.machine_area,
            simulation=DiagnosticSimulationElement(output_filename="end_screen.csrtrack"),
            physical=lastelem.physical,
        )
        csrtrackstr += lastscreen.to_csrtrack(n=counter["screen"])
        csrtrackstr += "}\n"
        self.csrtrack_headers["tracker"].end_time_marker = "screen" + str(counter["screen"]) + "b"
        for h in self.csrtrack_headers.values():
            csrtrackstr += h.write_CSRTrack()
        return csrtrackstr

    def to_wake_t(self) -> "Beamline":
        """
        Create a Wake-T-compatible beamline object based on the lattice information.

        Returns
        -------
        Segment
            A Wake-T `Beamline` object.
        """
        from wake_t import Beamline
        section_with_drifts = self.createDrifts()
        elem_dict = translate_elements(
            section_with_drifts.values(),
            master_lattice_location=self.master_lattice_location,
            directory=self.directory,
        )
        beamline = []
        for element in elem_dict.values():
            if not element.subelement:
                # try:
                if element.length > 0:
                    beamline.append(element.to_wake_t())
                # except Exception as e:
                #     print('Wake-T writeElements error:', element.name, e)
        return Beamline(beamline)


from ocelot.cpbd.elements import (
    Solenoid,
    SBend,
    Quadrupole,
    Sextupole,
    Octupole,
    Monitor,
    Marker,
    Cavity,
    Hcor,
    Undulator,
    TDCavity,
    Aperture,
    Drift,
)

ocelot_conversion_rules = {
    "Dipole": SBend,
    "Solenoid": Solenoid,
    "Quadrupole": Quadrupole,
    "Sextupole": Sextupole,
    "Octupole": Octupole,
    "Beam_Position_Monitor": Marker,
    "Beam_Arrival_Monitor": Drift,
    "Bunch_Length_Monitor": Drift,
    "Screen": Marker,
    "Rcollimator": Aperture,
    "Collimator": Aperture,
    "Monitor": Monitor,
    "Wall_Current_Monitor": Drift,
    "Integrated_Current_Transformer": Drift,
    "Faraday_Cup_Monitor": Drift,
    "RFDeflectingCavity": TDCavity,
    "RFCavity": Cavity,
    "Aperture": Aperture,
    "Shutter": Drift,
    "Valve": Drift,
    "Bellows": Drift,
    "Cleaner": Drift,
    "lscrift": Drift,
    "csrdrift": Drift,
    "edrift": Drift,
    "Drift": Drift,
    "Kicker": Hcor,
    "Scatter": Marker,
    "APContour": Marker,
    "Center": Marker,
    "FEL_Modulator": Undulator,
    "Wiggler": Undulator,
    "Charge": Marker,
    "Wakefield": Marker,
    "Watch_Point": Monitor,
    "Combined_Corrector": Hcor,
    "Marker": Marker,
    "Laser": Drift,
    "Plasma": Drift,
}



import os
import numpy as np
from typing import List, Dict, Any, Union
from pydantic import field_validator, BaseModel, ValidationInfo, Field, PositiveInt
from warnings import warn
from ._functions import read_yaml, merge_two_dicts
from .element import baseElement, Drift, PhysicalBaseElement, Diagnostic
from .physical import PhysicalElement, Position
from .baseModels import ModelBase
from .exceptions import LatticeError
import warnings

from .simulation import DriftSimulationElement


def dot(a, b) -> float:
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def chunks(li, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(li), n):
        yield li[i: i + n]

class BaseLatticeModel(ModelBase):
    """
    Base-level description for defining lattices. Allows dynamic extensibility via `append`, `remove` functions.

    This class should not be used for creating lattices from scratch; rather, use
    `nala.models.elementList.SectionLattice`, `nala.models.elementList.MachineLayout`.
    """

    name: str
    """Name of lattice model."""

    _basename: str

    master_lattice_location: str | None = None
    """Top-level directory containing lattice files."""

    def __add__(self, other: dict) -> dict:
        copy = getattr(self, self._basename).copy()
        copy.extend(other)
        return copy

    def __radd__(self, other: dict) -> dict:
        copy = other.copy()
        copy.extend(getattr(self, self._basename))
        return copy

    def __sub__(self, other):
        copy = getattr(self, self._basename).copy()
        if other in copy:
            del copy[other]
        return copy

    def append(self, other: Any) -> None:
        if not isinstance(other, list):
            other = [other]
        super().__init__(name=self.name, elements=self + other)
        setattr(self, self._basename, self + other)

    def remove(self, other: Any) -> None:
        if other in getattr(self, self._basename):
            copy = getattr(self, self._basename).copy()
            copy.remove(other)
            super().__init__(name=self.name, elements=copy)
            getattr(self, self._basename).remove(other)

    def __str__(self):
        return str({k: v.names() for k, v in getattr(self, self._basename).items()})

    def __repr__(self):
        return self.__str__()


class ElementList(ModelBase):
    """
    A container for an unordered dictionary of :class:`~nala.models.element.baseElement`.
    """
    elements: Dict[str, Union[baseElement, None]]

    def __str__(self):
        return str([e.name for e in self.elements.values()])

    def __getitem__(self, item: str) -> int:
        return self.elements[item]

    @property
    def names(self) -> list:
        return [e.name for e in self.elements.values()]

    def index(self, element: Union[str, baseElement]):
        if isinstance(element, str):
            return list(self.elements.keys()).index(element)
        return list(self.elements.values()).index(element)

    def _get_attributes_or_none(self, a):
        data = {}
        for k, v in self.elements.items():
            if v is not None and hasattr(v, a):
                data.update({k: getattr(v, a)})
            else:
                data.update({k: None})
        return data

    def __getattr__(self, a):
        try:
            return super().__getattr__(a)
        except Exception:
            data = self._get_attributes_or_none(a)
            if all([isinstance(d, (Union[baseElement, None])) for d in data.values()]):
                return ElementList(elements=data)
            return data

    def list(self):
        return list(self.elements.values())


class SectionLattice(BaseLatticeModel):
    """
    A section of a lattice, consisting of a list of elements and their order along the beam path.
    """

    order: List[str]
    """Ordered list of element names."""

    elements: ElementList = Field(default_factory=ElementList)
    """Container for elements."""

    # other_elements: ElementList = ElementList(elements={})
    # TODO should we put this back in?

    _basename: str = "elements"

    @field_validator("elements", mode="before")
    @classmethod
    def validate_elements(
        cls, elements: Union[List[baseElement], ElementList], info: ValidationInfo
    ) -> ElementList:
        if isinstance(elements, list):
            elemdict = {e.name: e for e in elements}
            # print([e for e in info.data['order'] if e not in elemdict.keys()])
            return ElementList(
                elements={
                    e: elemdict[e] for e in info.data["order"] if e in elemdict.keys()
                }
            )
        assert isinstance(elements, ElementList)
        return elements
    #
    # @model_serializer(mode="plain")
    # def serialize(self) -> dict:
    #     data = self.__dict__.copy()
    #     data['elements'] = {"elements": {}}
    #     data['elements']["elements"] = {
    #         k: v.model_dump() for k, v in self.elements.elements.items()
    #     }
    #     return data

    @property
    def names(self) -> List:
        """List of element names.

        Returns
        -------
        List
            List of element names.
        """
        return self.elements.names

    def __str__(self):
        # return str(getattr(self, self._basename).__str__())
        return str(self.names)

    def __getitem__(self, item: Union[str, int]) -> BaseModel:
        if isinstance(item, int):
            return self.elements[self.names[item]]
        return self.elements[item]

    def __getattr__(self, a):
        try:
            return super().__getattr__(a)
        except Exception:
            return getattr(self.elements, a)

    def _get_all_elements(self) -> List:
        """
        Get a list of all the elements in order.

        Returns
        -------
        List
            Ordered list of elements.
        """
        return [self.elements[e] for e in self.order if e in self.elements.names]

    def createDrifts(self, csr_enable: bool=True, lsc_enable: bool=True, lsc_bins: PositiveInt=20):
        """Insert drifts into a sequence of 'elements'"""
        positions = []
        originalelements = dict()
        elementno = 0
        newelements = dict()

        elements = self._get_all_elements()

        # if any([x != y for x, y in zip(elements[0].physical.start.model_dump(), [0, 0, 0])]):
        #     machine_area = elements[0].machine_area
        #     self.order.insert(0, "initial_marker")
        #     self.elements.elements.update(
        #         {
        #             "initial_marker": PhysicalBaseElement(
        #                 name="initial_marker",
        #                 hardware_class="Marker",
        #                 hardware_type="Marker",
        #                 machine_area=machine_area,
        #             )
        #         }
        #     )
        #     elements = self._get_all_elements()

        for elem in elements:
            if not elem.subelement:
                originalelements[elem.name] = elem
                if isinstance(elem, Diagnostic):
                    elem.physical.length = 0
                start = elem.physical.start.array
                end = elem.physical.end.array
                positions.append(start)
                positions.append(end)
        positions = positions[1:]
        positions.append(positions[-1])
        driftdata = list(
            zip(iter(list(originalelements.items())), list(chunks(positions, 2)))
        )

        for e, d in driftdata:
            newelements[e[0]] = e[1]
            if len(d) > 1:
                x1, y1, z1 = d[0]
                x2, y2, z2 = d[1]
                try:
                    length = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2)
                    vector = dot((d[1] - d[0]), [0, 0, 1])
                except Exception as exc:
                    print("Element with error = ", e[0])
                    print(d)
                    raise exc
                if round(length, 6) > 0:
                    elementno += 1
                    name = self.name + "_drift_" + str(elementno)
                    x, y, z = [(a + b) / 2.0 for a, b in zip(d[0], d[1])]
                    newdrift = Drift(
                        name=name,
                        machine_area=newelements[e[0]].machine_area,
                        hardware_class="drift",
                        physical=PhysicalElement(
                            length=abs(round(np.copysign(length, vector), 6)),
                            middle=Position(x=x, y=y, z=z),
                            datum=Position(x=x, y=y, z=z),
                        ),
                        simulation=DriftSimulationElement(
                            csr_enable=csr_enable,
                            lsc_enable=lsc_enable,
                            lsc_bins=lsc_bins,
                        )
                    )
                    newelements[name] = newdrift
        return newelements

    def get_s_values(
        self,
        as_dict: bool = False,
        at_entrance: bool = False,
        starting_s: float = 0
    ) -> list | dict:
        """
        Get the S values for the elements in the lattice.
        This method calculates the cumulative length of the elements in the lattice,
        starting from the entrance or the first element, depending on the `at_entrance` parameter.
        It returns a list or dict of S values, which represent the positions of the elements along the lattice.

        Parameters
        ----------
        as_dict: bool, optional
            If True, returns a dictionary with element names as keys and their S values as values.
        at_entrance: bool, optional
            If True, calculates S values starting from the entrance of the lattice.
            If False, calculates S values starting from the first element.
        starting_s: float, optional
            Initial s position

        Returns
        -------
        list | dict
            A list or dictionary of S values for the elements in the lattice.
            If `as_dict` is True, returns a dictionary with element names as keys and their S values as values.
            If `as_dict` is False, returns a list of S values.
        """
        elems = self.createDrifts()
        s = [starting_s]
        for e in list(elems.values()):
            s.append(s[-1] + e.physical.length)
        s = s[:-1] if at_entrance else s[1:]
        if as_dict:
            return dict(zip([e.name for e in elems.values()], s))
        return list(s)


class MachineLayout(BaseLatticeModel):
    """
    A machine layout, consisting of a dictionary of lattice sections.
    This class could represent a full beam path, for example.
    """

    sections: Dict[str, SectionLattice]  # = Field(frozen=True)
    """Dictionary of :class:`~nala.models.elementList.SectionLattice`, keyed by name."""

    master_lattice_location: str | None = None
    """Directory containing lattice files. """

    _basename: str = "sections"

    def model_post_init(self, __context):
        matrix = [v.elements.elements.values() for v in self.sections.values()]
        all_elems = [item for row in matrix for item in row]
        if len(all_elems) > 0:
            all_elems_reversed = reversed(all_elems)
            superelem = all_elems[-1].name
            start_pos = all_elems[-1].physical.start
            all_elem_corrected = []
            for elem in all_elems_reversed:
                vector = (
                    not elem.physical.end.vector_angle(start_pos, [0, 0, -1]) < -5e-6
                )
                if not elem.is_subelement():
                    superelem = elem.name
                subelem = (
                    elem.subelement == superelem if elem.is_subelement() else False
                )
                # if vector:
                all_elem_corrected += [elem]
                start_pos = elem.physical.start
            self._all_elements = list(reversed(all_elem_corrected))
        else:
            self._all_elements = {}

    @property
    def names(self) -> List:
        """
        Names of lattice sections

        Returns
        -------
        List
            Names of :attr:`~sections`
        """
        return list(self.sections.keys())

    def __str__(self):
        return str([k for k, v in self.sections.items()])

    def __getattr__(self, item: str):
        return getattr(self.sections, item)

    def __getitem__(self, item: str) -> int:
        return self.sections[item]

    def _get_all_elements(self) -> List[baseElement]:
        """
        List of all elements defined in the layout

        Returns
        -------
        List[baseElement]
            List of all elements.
        """
        return self._all_elements

    def _get_all_element_names(self) -> List[str]:
        """
        List of all element names defined in the layout.

        Returns
        -------
        List[str]
            Names of all elements.
        """
        return [e.name for e in self._get_all_elements()]

    def get_element(self, name: str) -> baseElement:
        """
        Return the LatticeElement object corresponding to a given machine element

        :param str name: Name of the element to look up
        :returns: :class:`~nala.models.element.baseElement` instance for that element
        """
        if name in self._get_all_element_names():
            index = self._get_all_element_names().index(name)
            return self._get_all_elements()[index]
        else:
            message = "Element %s does not exist along the beam path" % name
            raise LatticeError(message)

    def _get_element_names(self, lattice: list) -> list:
        """
        Return the name for each LatticeElement object in a list defining a lattice

        :param str lattice: List of LatticeElement objects representing machine hardware
        :returns: List of strings defining the names of the machine elements
        """
        return [ele.name for ele in lattice]

    def _lookup_index(self, name: str) -> int:
        """
        Look up the index of an element in a given lattice

        :param str name: Name of the element to search for
        :returns: List index of the item within that beam path
        """
        try:
            # fetch the index of the element
            return self._get_all_element_names().index(name)
        except ValueError:
            message = "Element %s does not exist along the beam path" % name
            raise LatticeError(message)

    @property
    def elements(self) -> List[str]:
        """
        List of all element names.

        Returns
        -------
        List[str]
            List of all element names.
        """
        return self._get_all_element_names()

    def _filter_element_list(self, result, filt, attrib):
        if isinstance(filt, (str, list)):
            # make list of valid types
            if isinstance(filt, str):
                filter_list = [filt.lower()]
            elif isinstance(filt, list):
                filter_list = [_type.lower() for _type in filt]
            # apply search criteria
            return [
                ele
                for ele in result
                if (
                    hasattr(ele, attrib) and getattr(ele, attrib).lower() in filter_list
                )
            ]
        return result

    def get_all_elements(
        self,
        element_type: Union[str, list, None] = None,
        element_model: Union[str, list, None] = None,
        element_class: Union[str, list, None] = None,
    ) -> List[str]:
        """
        Get all elements in the lattice, or filter them by type/model/class
        # TODO function name implies this returns elements rather than names; rename?

        Parameters
        ----------
        element_type: str | list | None
            Filter by element type; if list, gather multiple types; if None, gather all.
        element_model: str | list | None
            Filter by element model; if list, gather multiple models; if None, gather all.
        element_class
            Filter by element hardware class; if list, gather multiple classes; if None, gather all.

        Returns
        -------
        List[str]
            Filtered names of elements.
        """
        return self.elements_between(
            end=None,
            start=None,
            element_type=element_type,
            element_class=element_class,
            element_model=element_model,
        )

    def elements_between(
        self,
        end: str = None,
        start: str = None,
        element_type: Union[str, list, None] = None,
        element_model: Union[str, list, None] = None,
        element_class: Union[str, list, None] = None,
    ) -> List[str]:
        """
        Returns a list of all lattice elements (of a specified type) between
        any two points along the accelerator (inclusive). Elements are ordered according
        to their longitudinal position along the beam path.

        Parameters
        ----------
        end: str
            Name of the element defining the end of the search region
        start: str
            Name of the element defining the start of the search region
        element_type: str | list | None
            Filter by element type; if list, gather multiple types; if None, gather all.
        element_model: str | list | None
            Filter by element model; if list, gather multiple models; if None, gather all.
        element_class
            Filter by element hardware class; if list, gather multiple classes; if None, gather all.

        Returns
        -------
        List[str]
            Filtered names of elements.
        """
        # replace blank start and/or end point
        element_names = self._get_all_element_names()
        if start is None:
            start = element_names[0]
        if end is None:
            end = element_names[-1]

        # truncate the list between the start and end elements
        first = self._lookup_index(start)
        last = self._lookup_index(end) + 1
        result = self._get_all_elements()[first:last]

        result = self._filter_element_list(result, element_type, "hardware_type")
        result = self._filter_element_list(result, element_model, "hardware_model")
        result = self._filter_element_list(result, element_class, "hardware_class")

        return self._get_element_names(result)


class MachineModel(ModelBase):
    """
    The full model of the accelerator. It describes all :class:`~nala.models.elementList.MachineLayout` and
    :class:`~nala.models.elementList.SectionLattice` that particles can follow.
    These layouts and sections are also defined as Dict[str, list] and Dict[str, list], and the full dictionary
    containing all elements is also accessible.
    """

    layout: str | Dict | None = None
    """Dictionary containing layout names and the names of the sections of which they are composed."""

    section: str | Dict[str, Dict] | None = None
    """Dictionary containing section names and the elements that compose it."""

    elements: Dict[str, baseElement] = {}
    """Dictionary containing all elements defined in the machine model."""

    sections: Dict[str, SectionLattice] = {}
    """Dictionary containing :class:`~nala.models.elementList.SectionLattice`, keyed by name."""

    lattices: Dict[str, MachineLayout] = {}
    """Dictionary containing :class:`~nala.models.elementList.MachineLayout`, keyed by name.
    #TODO rationalise either this name `lattices` or the class name `MachineLayout`.
    """

    master_lattice_location: str | None = None
    """Directory containing lattice YAML files."""

    _layouts: List[str] = None

    _section_definitions: Dict = {}

    _default_path: str = None

    @field_validator("layout", mode="before")
    @classmethod
    def validate_layout(cls, v: str | dict) -> str | dict:
        if isinstance(v, str):
            if os.path.isfile(v):
                return v
            elif os.path.isfile(
                os.path.abspath(os.path.dirname(__file__) + "/../" + v)
            ):
                return os.path.abspath(os.path.dirname(__file__) + "/../" + v)
            else:
                raise ValueError(f"Directory {v} does not exist")
        elif isinstance(v, dict):
            if "layouts" not in v:
                raise KeyError("layout must specify lines each with a list of sections")
            return v
        else:
            raise ValueError("layout must be a str or dict")

    @field_validator("section", mode="before")
    @classmethod
    def validate_section(cls, v: str | dict) -> str | dict:
        if isinstance(v, str):
            if os.path.isfile(v):
                return v
            elif os.path.isfile(
                os.path.abspath(os.path.dirname(__file__) + "/../" + v)
            ):
                return os.path.abspath(os.path.dirname(__file__) + "/../" + v)
            else:
                raise ValueError(f"Directory {v} does not exist")
        elif isinstance(v, dict):
            if "sections" not in v:
                raise KeyError(
                    "section must specify sections each with lists of elements"
                )
            return v
        else:
            warnings.warn(
                "No sections specified. Sections will be generated from elements."
            )

    def model_post_init(self, __context):
        if isinstance(self.layout, str):
            config = read_yaml(self.layout)
            self._layouts = config.layouts
            try:
                self._default_path = config.default_layout
            except AttributeError:
                message = 'Missing "default_layout" in %s ' % self.layout_file
                warn(message)
        elif self.layout is None:
            self._layouts = {}
            self._default_path = None
            warnings.warn("No layouts specified. Lattices will be empty.")
        else:
            for key in ["layouts"]:
                if key not in self.layout:
                    raise KeyError("layout must specify layouts")
            self._layouts = self.layout["layouts"]
            if "default_layout" in self.layout:
                self._default_path = self.layout["default_layout"]
        if isinstance(self.section, str):
            config = read_yaml(self.section)
            self._section_definitions = config.sections
        elif self.section is None:
            self._section_definitions = {}
        else:
            if "sections" not in self.section:
                raise KeyError("section must specify sections with a list of sections")
            self._section_definitions = self.section["sections"]
        if len(self.elements) > 0:
            if self.section:
                self._build_layouts(self.elements)
            else:
                self._build_sections_from_elements(self.elements)

    def __add__(self, other) -> dict:
        copy = self.elements.copy()
        copy.update(other)
        return copy

    def __radd__(self, other) -> dict:
        copy = other.copy()
        copy.update(self.elements)
        return copy

    def __iter__(self) -> iter:
        return iter(self.elements)

    def __str__(self):
        return str(list(self.elements.keys()))

    def append(self, values: dict) -> None:
        self.elements = merge_two_dicts(values, self.elements)
        self._build_sections_from_elements(self.elements)
        self._build_layouts(self.elements)

    def update(self, values: dict) -> None:
        return self.append(values)

    def __getitem__(
        self, item: str | list[str] | tuple[str]
    ) -> BaseModel | List[BaseModel]:
        if isinstance(item, (list, tuple)):
            return [self.elements[subitem] for subitem in item]
        return self.elements[item]

    def __setitem__(self, item: str, value: Any) -> None:
        super().__init__(elements=self + {item: value})

    @property
    def default_path(self) -> str:
        return self._default_path

    @default_path.setter
    def default_path(self, path: str):
        self._default_path = path

    def _build_sections_from_elements(self, elements: dict) -> None:
        """build sections from the elements if no section definition is provided"""
        # Build a unique list of machine areas from the elements
        areas = set()
        for elem in elements.values():
            area = (
                elem.get("machine_area")
                if isinstance(elem, dict)
                else getattr(
                    elem,
                    "machine_area",
                    None,
                )
            )
            if area is not None:
                areas.add(area)
        areas = list(areas)
        for area in areas:
            # collect list of elements from this machine area
            new_elements = [
                x
                for x in elements.values()
                if (
                    x.get("machine_area")
                    if isinstance(x, dict)
                    else getattr(x, "machine_area", None)
                )
                == area
            ]
            self.sections[area] = SectionLattice(
                name=area,
                elements=new_elements,
                order=[
                    e["name"] if isinstance(e, dict) else e.name for e in new_elements
                ],
                master_lattice_location=self.master_lattice_location,
            )
            if not self._section_definitions or area not in self._section_definitions:
                self._section_definitions[area] = [
                    e["name"] if isinstance(e, dict) else e.name for e in new_elements
                ]
        self.lattices = {}

    def _build_layouts(self, elements: dict) -> None:
        """build lists defining the lattice elements along each possible beam path"""
        # build dictionary with a lattice for each beam path
        if self._layouts:
            for path, areas in self._layouts.items():
                for _area in areas:
                    if _area in self._section_definitions:
                        # collect list of elements from this machine area
                        new_elements = [
                            x
                            for x in elements.values()
                            if x.name in self._section_definitions[_area]
                        ]
                        try:
                            self.sections[_area] = SectionLattice(
                                name=_area,
                                elements=new_elements,
                                order=self._section_definitions[_area],
                                master_lattice_location=self.master_lattice_location,
                            )
                        except KeyError:
                            pass
                    else:
                        print("MachineModel", "_build_layouts", _area, "missing")
                self.lattices[path] = MachineLayout(
                    name=path,
                    sections={
                        _area: self.sections[_area]
                        for _area in areas
                        if _area in self.sections
                    },
                    master_lattice_location=self.master_lattice_location,
                )
            if len(self.lattices) == 1 and self._default_path is None:
                self._default_path = list(self.lattices.keys())[0]
        else:
            for _area, elem_names in self._section_definitions.items():
                # collect list of elements from this machine area
                new_elements = [
                    x
                    for x in elements.values()
                    if (x.name in self._section_definitions[_area])
                ]
                self.sections[_area] = SectionLattice(
                    name=_area,
                    elements=new_elements,
                    order=elem_names,
                    master_lattice_location=self.master_lattice_location,
                )
            self.lattices = {}

    def get_element(self, name: str) -> baseElement:
        """
        Return the LatticeElement object corresponding to a given machine element

        :param str name: Name of the element to look up
        :returns: LatticeElement instance for that element
        """
        if name in self.elements:
            return self.elements[name]
        else:
            message = (
                "Element %s does not exist anywhere in the accelerator lattice" % name
            )
            raise LatticeError(message)

    def get_all_elements(
        self,
        element_type: Union[str, list, None] = None,
        element_model: Union[str, list, None] = None,
        element_class: Union[str, list, None] = None,
    ) -> List[str]:
        """
        Get all elements in the lattice, or filter them by type/model/class
        # TODO function name implies this returns elements rather than names; rename?

        Parameters
        ----------
        element_type: str | list | None
            Filter by element type; if list, gather multiple types; if None, gather all.
        element_model: str | list | None
            Filter by element model; if list, gather multiple models; if None, gather all.
        element_class
            Filter by element hardware class; if list, gather multiple classes; if None, gather all.

        Returns
        -------
        List[str]
            Filtered names of elements.
        """
        return self.elements_between(
            end=None,
            start=None,
            element_type=element_type,
            element_class=element_class,
            element_model=element_model,
        )

    def elements_between(
        self,
        end: str = None,
        start: str = None,
        element_type: Union[str, list, None] = None,
        element_model: Union[str, list, None] = None,
        element_class: Union[str, list, None] = None,
        path: str = None,
    ) -> List[str]:
        """
        Returns a list of all lattice elements (of a specified type) between
        any two points along the accelerator (inclusive). Elements are ordered according
        to their longitudinal position along the beam path.

        Parameters
        ----------
        end: str
            Name of the element defining the end of the search region
        start: str
            Name of the element defining the start of the search region
        element_type: str | list | None
            Filter by element type; if list, gather multiple types; if None, gather all.
        element_model: str | list | None
            Filter by element model; if list, gather multiple models; if None, gather all.
        element_class
            Filter by element hardware class; if list, gather multiple classes; if None, gather all.
        path: str
            Optional beam path, i.e. name of :class:`~nala.models.elementList.MachineLayout`.

        Returns
        -------
        List[str]
            Filtered names of elements.
        """
        # determine the beam path
        if path is None:
            if hasattr(self, "_default_path") and self._default_path in self.lattices:
                path = self._default_path
            else:
                raise Exception(
                    '"default_layout" = %s is not defined, and more than one layout exists.'
                    % self._default_path,
                )
        elif path not in self.lattices:
            raise Exception('"path" = %s is not defined' % path)

        if end is None:
            path_obj = self.lattices[path]
            end = path_obj.elements[-1]
        else:
            end_obj = self.get_element(end)
            beam_path = (
                end_obj.machine_area
                if (end_obj.machine_area in self.lattices)
                else path
            )
            path_obj = self.lattices[beam_path]

        # find the start of the search area
        if start is None:
            start = path_obj.elements[0]

        # return a list of elements along this beam path
        elements = path_obj.elements_between(
            start=start,
            end=end,
            element_type=element_type,
            element_class=element_class,
            element_model=element_model,
        )
        return elements



from pydantic import BaseModel, ConfigDict, Field
from typing import Dict, List, Any, Literal
from ...utils.classes import getGrids

class opal_header(BaseModel):
    """
    Generic class for generating OPAL namelists

    See `OPAL manual`_ for more details.

    .. OPAL manual: https://amas.web.psi.ch/opal/Documentation/master/OPAL_Manual.html
    """

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_assignment=True,
        populate_by_name=True,
    )

    objectname: str = Field(alias="name")
    """Name of the object, used as a unique identifier in the simulation."""

    objecttype: str = Field(alias="type")
    """Type of the object, which determines its behavior and properties in the simulation."""

    header: str
    """Name of OPAL header"""

    exclude: List[str] = [
        "objectname",
        "objecttype",
        "opaldict",
        "header",
        "exclude",
        "breakstr",
    ]

    opaldict: Dict = {}
    """Dictionary containing values to be written to the header"""

    breakstr: str = "//----------------------------------------------------------------------------"
    """String used for separating headers in the input file"""

    def write_Opal(self) -> str:
        """
        Write the text for the Opal namelist based on its attributes.

        Returns
        -------
        str
            Opal-compatible string representing the namelist
        """
        output = f"{self.breakstr}\n// {self.header}\n"
        if self.objecttype != "":
            output += f"{self.header}: {self.objecttype}, \n"
        else:
            output += f"{self.header}, \n"
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                if key in self.opaldict:
                    output += f"\t{self.opaldict[key]} = {val},\n"
                elif key == "METHOD":
                    output += "METHOD = \"PARALLEL-T\",\n"
                else:
                    output += f"\t{key} = {val},\n"
        output = output[:-2] + ";\n"
        return output

class opal_option(opal_header):
    """
    Class for generating the OPTION namelist for OPAL. See `OPAL manual`_ for more details.
    """

    objectname: str = "option"
    """Name of object"""

    objecttype: str = "OPTION"
    """Type of object"""

    header: str = "OPTION"
    """Name of header"""

    VERSION: str = "202210"
    """Used to indicate for which version of OPAL the input file is written.The major and minor versions of OPAL 
    and of the input file have to match.The patch version of OPAL must be greater or equal to the patch version 
    of the input file.If the version doesn’t fulfill above criteria OPAL stops immediately and prints instructions 
    on how to convert the input file.Starting with version 1.6.0 of OPAL the option VERSION is mandatory in the 
    OPAL input file.The format is Mmmpp where M stands for the major, m for the minor and p for the patch version.
    For version 1.6.0 of OPAL VERSION should read 10600."""

    AMR: bool = False
    """Enable adaptive mesh refinement. Its default value is false."""

    AMR_REGRID_FREQ: int = None
    """Defines after how many steps an AMR regrid is performed."""

    AMR_YT_DUMP_FREQ: int = None
    """The frequency to dump grid and particle data for AMR."""

    ASCIIDUMP: bool = False
    """If true, instead of HDF5, ASCII output is generated for the following elements: 
    Collimator, Geometry, Monitor, Probe, Stripper, Vacuum and global losses."""

    AUTOPHASE: int = 6
    """Defines how accurate the search for the phase at which the maximal energy is gained should be. 
    The higher this number the more accurate the phase will be. 
    If it is set to 0 then the auto-phasing algorithm isn’t run."""

    BEAMHALOBOUNDARY: float = None
    """Defines in terms of sigma where the halo starts. Only used in OPAL-cycl."""

    BOUNDPDESTROYFQ: int = None
    """The frequency to delete particles which are too far away from the center of beam. 
    Only used in OPAL-cycl."""

    CLOTUNEONLY: bool = False
    """If set to true, stop after closed orbit finder and tune calculation. 
    Only used in OPAL-cycl."""

    COMPUTEPERCENTILES: bool = False
    """If true, the 68 (1 sigmas for normal distribution), the 95 (2 sigmas), the 99.7 (3 sigmas) and 
    the 99.99 (4 sigmas) percentiles of the bunch size and the normalized emittance are calculated. 
    The data are stored into .stat output file and loss file in HDF5 format (.h5). 
    The calculation is performed whenever the number of particles exceeds 100."""

    CSRDUMP: bool = False
    """If true the electric csr field component, Ez, line density and the derivative of the line density 
    is written into the data directory. The first line gives the average position of the beam bunch. 
    Subsequent lines list zz position of longitudinal mesh (with respect to the head of the beam bunch), 
    Ez, line density and the derivative of the line density. Note that currently the line density derivative 
    needs to be scaled by the inverse of the mesh spacing to get the correct value. T
    he CSR field is dumped at each time step of the calculation. E
    ach text file is named 'Bend Name' (from input file) + '-CSRWake' + 'time step number in that bend 
    (starting from 1)' + '.txt'."""

    CZERO: bool = False
    """If true the distributions are generated such that the centroid is exactly zero and not 
    statistically dependent."""

    DELPARTFREQ: int = None
    """The frequency to delete particles in OPAL-cycl."""

    DUMPBEAMMATRIX: bool = None
    """If true, the 6-dimensional beam matrix (upper triangle only) is stored in the statistics output file (.stat)."""

    EBDUMP: bool = None
    """If true the electric and magnetic field on the particle is saved each time a phase space is written."""

    ENABLEHDF5: bool = None
    """If true (default), HDF5 read and write is enabled."""

    ENABLEVTK: bool = None
    """If true (default), VTK write of geometry voxel mesh is enabled."""

    HALOSHIFT: float = None
    """Constant parameter to shift halo value. Its default value is 0.0."""

    IDEALIZED: bool = None
    """Instructs to use the hard edge model for the calculation of the path length in OPAL-t. T
    he path length is computed to place the elements in the three-dimensional space from ELEMEDGE."""

    LOGBENDTRAJECTORY: bool = None
    """Save the reference trajectory inside dipoles in an ASCII file. 
    For each dipole a separate file is written to the directory data/."""

    MEMORYDUMP: bool = None
    """If true, it writes the memory consumption of every core to a SDDS file (.mem). 
    The write frequency corresponds to STATDUMPFREQ."""

    MINBINEMITTED: int = None
    """The number of bins that have to be emitted before the bin are squashed into a single bin."""

    MINSTEPFORREBIN: int = None
    """The number of steps into the simulation before the bins are squashed into a single bin. 
    This should be used instead of the global variable of the same name. Its default value is 200."""

    MTSSUBSTEPS: int = None
    """Only used for multiple-time-stepping (MTS) integrator in OPAL-cycl. Specifies how many sub-steps 
    for external field integration are done per step. Making less steps per turn and increasing this value 
    is the recommended way to reduce space charge solve frequency. Its default value is 1."""

    NLHS: int = None
    """Number of stored old solutions for extrapolating the new starting vector. 
    Default value is 1 and just the last solution is used."""

    NUMBLOCKS: int = None
    """Maximum number of vectors in the Krylov space for ITSOLVER=CG or ITSOLVER=GMRES (see FIELDSOLVER command). 
    Its default value is 0."""

    PSDUMPFREQ: int = None
    """Defines after how many time steps the phase space is dumped into the H5hut file (.h5). 
    It also controls the frequency of phase space printed on the standard output. Its default value is 10."""

    PSDUMPEACHTURN: bool = None
    """Control option of phase space dumping. If true, dump phase space after each turn. 
    For the time being, this is only use for multi-bunch simulation in OPAL-cycl. Its default value is false."""

    PSDUMPFRAME: str = None
    """Control option that defines the frame in which the phase space data is written for .h5 and .stat files. 
    Beware that the data is written in a given time step. Most accelerator physics quantities are defined at a 
    given s-step where s is distance along the reference trajectory. For non-isochronous accelerators, 
    particles at a given time step can be quite a long way away from the reference particle, 
    yielding unexpected results. Currently only available for OPAL-cycl. 
    In OPAL-t the phase-space is always written in the frame of the reference particle.

        GLOBAL: data is written in the global Cartesian frame;

        BUNCH_MEAN: data is written in the bunch mean frame or;

        REFERENCE: data is written in the frame of the reference particle."""

    REBINFREQ: int = None
    """Defines after how many time steps we update the energy Bin ID of each particle. For the time being. 
    Only available for multi-bunch simulation in OPAL-cycl. Its default value is 100."""

    RECYCLEBLOCKS: int = None
    """Number of vectors in the recycle space for ITSOLVER=CG or ITSOLVER=GMRES (see FIELDSOLVER command). 
    Its default value is 0."""

    REMOTEPARTDEL: float = None
    """Artificially delete remote particles if their distances to the beam centroid is larger than 
    REMOTEPARTDEL times of the beam rms size. In OPAL-t only the longitudinal component of the particles 
    is considered. In OPAL-cycl all components are considered if the the value is negative 
    (further on using its absolute value) and only the transverse components if the value is positive. 
    Its default value is 0.0, i.e. no particles are deleted."""

    REPARTFREQ: int = None
    """Defines after how many time steps we do particles repartition to balance the computational 
    load of the computer nodes. Its default value is 10."""

    RHODUMP: bool = None
    """If true the scalar ρρ field is saved each time a phase space is written. There exists a reader in 
    Visit with versions greater or equal 1.11.1. Its default value is false."""

    RNGTYPE: str = None
    """The name (see String Attributes) of a random number generator can be provided. 
    The default random number generator (RANDOM) is a portable 48-bit generator. 
    Three quasi random generators are available: HALTON, SOBOL and NIEDERREITER. 
    For details see the GSL reference manual (18.5)."""

    SCSOLVEFREQ: int = None
    """If the space charge field is slowly varying w.r.t. external fields, this option allows to change the 
    frequency of space charge calculation, i.e. the space charge forces are evaluated every SCSOLVEFREQ 
    step and then reused for the following steps. Affects integrators LF2 and RK4 of OPAL-cycl. 
    Its default value is 1. Note: as the multiple-time-stepping (MTS) integrator maintains accuracy much 
    better with reduced space charge solve frequency, this option should probably not be used anymore."""

    SEED: int = None
    """Selects a particular sequence of random values. A SEED value is an integer in the range [0…999999999] 
    (default: 123456789). SEED can be an expression. If SEED == -1, the time is used as seed and the generator 
    is not portable anymore. See also Deferred Expressions and Random Values."""

    SPTDUMPFREQ: int = None
    """Defines after how many time steps we dump the phase space of single particle in OPAL-cycl. 
    It is always useful to record the trajectory of reference particle or some specified particle for primary study. 
    Its default value is 1."""

    STATDUMPFREQ: int = None
    """Defines after how many time steps we dump statistical data, such as RMS beam emittance, to the .stat file. 
    Its default value is 10."""

    def write_Opal(self) -> str:
        """
        Write the text for the Opal namelist based on its attributes.

        Returns
        -------
        str
            Opal-compatible string representing the namelist
        """
        output = f"{self.breakstr}\n// {self.header}\n"
        for key, val in self.model_dump().items():
            if key not in self.exclude and val is not None:
                output += f"{self.header}, {key} = {val};\n"
        return output

class opal_distribution(opal_header):
    """
    Class for generating the OPTION namelist for OPAL. See `OPAL manual`_ for more details.

    Note that only FROMFILE distributions are currently supported.
    """

    header: str = "DIST"
    """Name of header"""

    objectname: str = "distribution"
    """Name of object"""

    objecttype: str = "DISTRIBUTION"
    """Type of object"""

    TYPE: Literal["FROMFILE"] = "FROMFILE"

    input_particle_definition: str

    def model_post_init(self, context: Any, /) -> None:
        self.opaldict = {"input_particle_definition": "FNAME"}

    def write_Opal(self) -> str:
        if not self.input_particle_definition:
            raise ValueError("input_particle_definition must be defined for opal_distribution")
        return super().write_Opal()

class opal_fieldsolver(opal_header):
    """
    Class for generating the FIELDSOLVER namelist for OPAL. See `OPAL manual`_ for more details.

    Note that the AMR solve is not currently supported; only FFT has really been tested.
    """

    objectname: str = "fieldsolver"
    """Name of object"""

    objecttype: str = "FIELDSOLVER"
    """Type of object"""

    header: str = "FS"
    """Name of header"""

    npart: int
    """Number of particles in the bunch"""

    space_charge_mode: str = "False"
    """Space charge mode"""

    grids: getGrids | None = None
    """Space charge grids"""

    sample_interval: int = 1
    """Downsampling interval calculated as 2 ** (3 * sample_interval)"""

    FSTYPE: Literal["FFT", "FFTPERIODIC", "SAAMG", "P3M", "NONE"] = "FFT"
    """Specify the type of field solver: FFT, FFTPERIODIC, SAAMG, P3M and NONE. 
    Further arguments are enabled with the AMR solver (cf. Adaptive Mesh Refinement (AMR) Solver)."""

    PARFFTX: bool = True
    """If TRUE, the dimension x is distributed among the processors"""

    PARFFTY: bool = True
    """If TRUE, the dimension y is distributed among the processors"""

    PARFFTT: bool = True
    """If TRUE, the dimension t is distributed among the processors"""

    MX: int = None
    """Number of grid points in x specifying rectangular grid"""

    MY: int = None
    """Number of grid points in y specifying rectangular grid"""

    MT: int = None
    """Number of grid points in t specifying rectangular grid"""

    BCFFTX: str = "open"
    """Boundary condition in x [OPEN] (FFT + AMR_MG only)."""

    BCFFTY: str = "open"
    """Boundary condition in y [OPEN] (FFT + AMR_MG only)."""

    BCFFTZ: str = "open"
    """Boundary condition in z [OPEN,PERIODIC] (FFT + AMR_MG only)."""

    GREENSF: str = "Integrated"
    """Defines the Greens function for the FFT-based solvers (FFT + P3M only)."""

    BBOXINCR: float = None
    """Enlargement of the bounding box in %."""

    ITSOLVER: str = None
    """Type of iterative solver (SAAMG + AMR_MG only)."""

    RC: float = None
    """Defines the cut-off radius in the boosted frame for the P3M solver (P3M only)."""

    ALPHA: float = None
    """Defines the interaction splitting parameter for the P3M solver with standard Green’s function 
    (P3M + GREENSF=STANDARD only)."""

    def model_post_init(self, context: Any, /) -> None:
        self.grids = getGrids()
        self.opaldict = {"input_particle_definition": "FNAME"}
        self.exclude.extend(["npart", "space_charge_mode", "grids", "sample_interval"])
        self.MX = self.grid_size
        self.MY = self.grid_size
        self.MT = self.grid_size
        if self.space_charge:
            self.FSTYPE = "FFT"
        else:
            self.FSTYPE = "NONE"

    def write_Opal(self) -> str:
        if not self.npart:
            raise ValueError("npart must be defined for opal_fieldsolver")
        return super().write_Opal()

    @property
    def space_charge(self) -> bool:
        """
        Flag to indicate whether space charge is enabled.

        Returns
        -------
        bool
            True if enabled
        """
        return not (
                self.space_charge_mode == "False"
                or self.space_charge_mode is False
                or self.space_charge_mode is None
                or self.space_charge_mode == "None"
        )

    @property
    def grid_size(self) -> int:
        """
        Get the number of space charge bins.

        Returns
        -------
        int
            The number of space charge bins based on the number of particles
        """
        # print('asking for grid sizes n = ', self.npart, ' is ', self.grids.getGridSizes(self.npart))
        return self.grids.getGridSizes(self.npart / self.sample_interval)

class opal_beam(opal_header):
    """
    Class for generating the BEAM namelist for OPAL. See `OPAL manual`_ for more details.

    Note that only electrons, positrons and protons are currently supported.
    """

    objectname: str = "beam"
    """Name of object"""

    objecttype: str = "BEAM"
    """Type of object"""

    header: str = "BEAM1"
    """Name of header"""

    PARTICLE: Literal["ELECTRON", "POSITRON", "PROTON"]
    """The name of particles in the machine."""

    PC: float
    """Particle momentum in GeV/c."""

    NPART: int
    """Number of particles."""

    CHARGE: int
    """The particle charge expressed in elementary charges."""

    BFREQ: int = 1
    """The bunch frequency in MHz."""

    BCURRENT: float
    """The bunch current in A. BCURRENT=Q×BFREQBCURRENT=Q×BFREQ with Q the total charge.
    So essentially this is set to the charge of the bunch in micro-coulombs."""


class opal_track(opal_header):
    """
    Class for generating the TRACK namelist for OPAL. See `OPAL manual`_ for more details.
    """

    objectname: str = "track"
    """Name of object"""

    objecttype: str = ""
    """Type of object"""

    header: str = "TRACK"
    """Name of header"""

    LINE: str
    """The label of a preceding LINE (no default)."""

    BEAM: str = "BEAM1"
    """The named BEAM command defines the particle mass, charge and reference momentum.
    This should be the same as the name provided by opal_beam."""

    T0: float = None
    """The initial time [s] of the simulation, its default value is 0."""

    DT: float | str = 1e-12
    """Array of time step sizes for tracking, default length of the array is 1 and its only value is 1 ps."""

    MAXSTEPS: int = None
    """Array of maximal number of time steps, default length of the array is 1 and its only value is 10."""

    ZSTART: float = None
    """Initial position of the reference particle along the reference trajectory, default position is 0.0 m."""

    ZSTOP: float | str
    """Array of z-locations [m], default length of the array is 1 and its only value is 1E61E6 [m]. 
    The simulation switches to the next set, i+1i+1, of DT, MAXSTEPS and ZSTOP if either it has been t
    racking with the current set for more than MAXSTEPS steps or the mean position has reached a z-position 
    larger than ZSTOP. If set i is the last set of the array then the simulation stops."""

    TIMEINTEGRATOR: Literal["RK4", "LF2", "MTS"] = None
    """Define the time integrator. Currently only available in OPAL-cycl. The valid options are RK4, LF2 and MTS"""

    def write_Opal(self) -> str:
        self.DT = str(self.DT)
        self.ZSTOP = "{" + str(self.ZSTOP + 1e-1) + "}"
        return super().write_Opal()

class opal_run(opal_header):
    """
    Class for generating the RUN namelist for OPAL. See `OPAL manual`_ for more details.

    Note that only OPAL-t is currently supported.
    """

    objectname: str = "run"
    """Name of object"""

    objecttype: str = ""
    """Type of object"""

    header: str = "RUN"
    """Name of header"""

    METHOD: str = "PARALLEL-T"
    """The name (a string, see String Attributes) of the tracking method to be used. 
    For the time being the following methods are known:

        PARALLEL-T This method puts OPAL in OPAL-t mode (see Chapter OPAL-t).

        CYCLOTRON-T This method puts OPAL in OPAL-cycl mode (see Chapter OPAL-cycl)."""

    FIELDSOLVER: str = "FS"
    """The field solver to be used. This should be the same as the name provided by opal_fieldsolver"""

    DISTRIBUTION: str = "DIST"
    """The particle distribution to be used. This should be the same as the name provided by opal_distribution"""

    BEAM: str = "BEAM1"
    """The particle beam to be used. This should be the same as the name provided by opal_beam"""

    TURNS: int = 1
    """The number of turns (integer) to be tracked (default: 1, namely single bunch).

    In OPAL-cycl, this parameter represents the number of bunches that will be injected into the cyclotron. 
    In restart mode, the code firstly reads an attribute NumBunch from .h5 file which records how many bunches 
    have already been injected. If NumBunch << TURNS, the last TURNS −− NumBunch bunches will be injected 
    in sequence by reading the initial distribution from the .h5 file."""

    MBMODE: Literal["AUTO", "FORCE"] = None
    """This defines which mode of multi-bunch runs. There are two options for it, namely, AUTO and FORCE. 
    See Multi-bunch Mode for their explanations in detail.
    For restarting run with TURNS larger than one, if the existing bunches of the read-in step is larger than one, 
    the mode is forcedly set to FORCE. Otherwise, it is forcedly set to AUTO.
    This argument is available for OPAL-cycl."""

    PARAMB: float = None
    """This is a control parameter to define when to start to transfer from single bunch to multi-bunches for 
    AUTO mode (default: 5.0). This argument is only available for AUTO mode multi-bunch run in OPAL-cycl."""

    MB_BINNING: Literal["GAMMA_BINNING", "BUNCH_BINNING"] = None
    """Type of energy binning in multi-bunch mode: GAMMA_BINNING or BUNCH_BINNING (default: GAMMA_BINNING). 
    When BUNCH_BINNING binning, then all particles of a bunch are in the same energy bin. When GAMMA_BINNING 
    binning, then the bin depends on the momentum of the particle. Only available in OPAL-cycl."""

    MB_ETA: float = None
    """The scale parameter for binning in multi-bunch mode (default: 0.01). Only used in MB_BINNING=GAMMA_BINNING. 
    Only available in OPAL-cycl."""

    TRACKBACK: bool = None
    """The particles are tracked backward in time if TRACKBACK=TRUE. OPAL starts at ZSTART and tracks the 
    bunch back in time. It changes the size of the time step when it crosses the thresholds given in the 
    ZSTOP attribute of the TRACK command and stops once it reaches the lowest item of ZSTOP. 
    Only available in OPAL-t. Default is FALSE."""



import yaml
from yaml import CSafeLoader as Loader

from ..models.element import *  # noqa


def interpret_YAML_Element(elem):
    if "hardware_type" in elem and elem["hardware_type"] in globals():
        try:
            felem = globals()[elem["hardware_type"]]
            elemmodel = felem(**elem)
            return elemmodel
        except Exception as e:
            print("interpret_YAML_Element - Error", e)


def read_YAML_Element_File(filename):
    with open(filename, "r") as stream:
        data = yaml.load(stream, Loader=Loader)
    return interpret_YAML_Element(data)


def read_YAML_Element_Files(filenames: list):
    data = ""
    for file in filenames:
        data += "\n---\n"
        with open(file, "r") as stream:
            data += stream.read()
    gen = list(yaml.load_all(data, Loader=Loader))
    return gen


def read_YAML_Combined_File(filename):
    with open(filename, "r") as stream:
        elements = yaml.load(stream, Loader=Loader)
    return [interpret_YAML_Element(element) for element in elements.values()]



import numpy as np
from scipy.constants import speed_of_light, pi
from pydantic import (
    BaseModel,
    model_serializer,
    Field,
    field_validator,
    NonNegativeInt,
    create_model,
    NonNegativeFloat,
    computed_field,
)
from typing import Dict, Any, List, Union
from .baseModels import IgnoreExtra, T


def Power(a, b):
    return a**b


def Sqrt(a):
    return Power(a, 0.5)


Pi = pi
Degree = pi / 180.0


class Multipole(BaseModel):
    """
    Single order magnetic multipole model.
    """

    order: NonNegativeInt = 0
    """Multipole order (0=dipole, 1=quadrupole, etc.)."""

    normal: float = 0.0
    """Normal component of the multipole strength."""

    skew: float = 0.0
    """Skew component of the multipole strength."""

    radius: float = 0.0
    """Reference radius for the multipole strength."""


multipoles = {
    "K" + str(no) + "L": (Multipole, Field(default=Multipole(order=no), repr=False))
    for no in range(0, 13)
}
MultipolesData = create_model("Multipoles", **multipoles)


class Multipoles(MultipolesData):
    """
    Magnetic multipoles model.
    """

    @field_validator("*", mode="before")
    def validate_Multipole(cls, v: Union[List, dict]) -> Multipole:
        if isinstance(v, (list, tuple)):
            if len(v) == 2:
                return Multipole(order=v[0], normal=v[1])
            elif len(v) == 4:
                return Multipole(order=v[0], normal=v[1], skew=v[2], radius=v[3])
        elif isinstance(v, (dict)):
            return Multipole(**v)
        elif isinstance(v, (Multipole)):
            return v
        else:
            raise ValueError("Multipole should be a dict or a list of floats")

    # def __str__(self):
    #     return " ".join(
    #         [
    #             "K"
    #             + str(i)
    #             + "L=Multipole("
    #             + getattr(self, "K" + str(i) + "L").__str__()
    #             + ")"
    #             for i in range(0, 13)
    #             if abs(getattr(self, "K" + str(i) + "L").normal) > 0
    #             or abs(getattr(self, "K" + str(i) + "L").skew) > 0
    #         ]
    #     )

    def __repr__(self):
        return "Multipoles(" + self.__str__() + ")"

    @model_serializer
    def ser_model(self) -> Dict[str, Any]:
        cls = self.__class__
        return {
            k: getattr(self, k)
            for k in cls.model_fields.keys()
            if abs(getattr(self, k).normal) > 0 or abs(getattr(self, k).skew) > 0
        }

    def normal(self, order: int) -> Union[int, float]:
        """
        Get the normal component of the multipole strength for a given order.

        Args:
            order (int): The order of the multipole (0=dipole, 1=quadrupole, etc.).

        Returns:
            Union[int, float]: The normal component of the multipole strength.
        """
        return getattr(self, "K" + str(order) + "L").normal

    def skew(self, order: int) -> Union[int, float]:
        """
        Get the skew component of the multipole strength for a given order.

        Args:
            order (int): The order of the multipole (0=dipole, 1=quadrupole, etc.).

        Returns:
            Union[int, float]: The skew component of the multipole strength.
        """
        return getattr(self, "K" + str(order) + "L").skew

    def __eq__(self, other) -> bool:
        return self.ser_model() == other

    def __neq__(self, other) -> bool:
        return not self.__eq__(other)


class FieldIntegral(BaseModel):
    """
    Field integral coefficients model.
    """

    coefficients: List[Union[int, float]] = [0]
    """Integrated field coefficients."""

    def currentToK(self, current: float, energy: float) -> float:
        """
        Convert the current in the magnet to the normalized strength (K value).
        The method calculates the normalized strength (K value) of the magnetic field
        based on the provided current and energy. It uses the field integral coefficients
        to compute the integrated field strength and applies a scaling factor based on
        the speed of light and the beam energy.

        Args:
            current (float): The current flowing through the magnet (in amperes).
            energy (float): The energy of the particle beam (in MeV).

        Returns:
            float: The normalized strength (K value) of the magnetic field.
        """
        sign = np.copysign(1, current)
        ficmod = [i * int(sign) for i in self.coefficients[:-1]]
        coeffs = np.append(ficmod, self.coefficients[-1])
        int_strength = np.polyval(coeffs, abs(current))
        effect = (speed_of_light / 1e6) * int_strength / energy
        return effect

    def __iter__(self) -> iter:
        return iter(self.coefficients)


class LinearSaturationFit(BaseModel):
    """
    Linear + saturation fit coefficients model.
    """

    m: float = 0
    I_max: NonNegativeFloat = 0
    f: float = 0
    a: float = 0
    I0: float = 0
    d: float = 0
    L: NonNegativeFloat = 0

    @property
    def coefficients(self) -> List[Union[int, float]]:
        return [self.m, self.I_max, self.f, self.a, self.I0, self.d, self.L]

    @classmethod
    def from_string(cls, v: Union[str, List]) -> T:
        if isinstance(v, str):
            coeff_list = list(map(float, v.strip().split(",")))
            assert len(coeff_list) == len(cls.model_fields.keys())
            return cls(**{k: v for k, v in zip(cls.model_fields.keys(), coeff_list)})
        elif isinstance(v, (list, tuple)):
            assert len(v) == len(cls.model_fields.keys())
            return cls(**{k: v for k, v in zip(cls.model_fields.keys(), v)})
        else:
            raise ValueError(
                "LinearSaturationFit should be a string or a list of floats"
            )

    def update_from_string(self, v: Union[str, List]) -> None:
        cls = self.__class__
        if isinstance(v, str):
            coeff_list = list(map(float, v.strip().split(",")))
            assert len(coeff_list) == len(cls.model_fields.keys())
            [setattr(self, k, v) for k, v in zip(cls.model_fields.keys(), coeff_list)]
        elif isinstance(v, (list, tuple)):
            assert len(v) == len(cls.model_fields.keys())
            [setattr(self, k, v) for k, v in zip(cls.model_fields.keys(), v)]

    def currentToK(self, current: float, momentum: float | None = None) -> Dict:
        """
        Convert the current in the magnet to the normalized strength (K value).

        The method calculates the normalized strength (K value) of the magnetic field
        based on the provided current and momentum. It uses the field integral coefficients
        to compute the integrated field strength and applies a scaling factor based on
        the speed of light and the beam momentum.

        Args:
            current (float): The current flowing through the magnet (in amperes).
            momentum (float): The momentum of the particle beam (in MeV/c).

        Returns:
            dict: A dictionary containing the K value, KL value, gradient, and integrated strength.
                The K value is the normalized strength of the magnetic field, KL is the K value multiplied by the
                length of the magnet, gradient is the magnetic field gradient, and integrated strength is the
                integrated field strength.
        """
        abs_I = abs(current)
        m, I_max, f, a, I0, d, L = list(self.coefficients)
        int_strength = (
            m * current
            if abs_I < I_max
            else np.copysign((f * abs_I**3 + a * (abs_I - I0) ** 2 + d), current)
        ) / 1000
        gradient = int_strength / L
        if momentum is not None:
            KL = 1 * (speed_of_light / 1e6) * int_strength / momentum
            return {
                "K": KL * 1000 / L,
                "KL": KL,
                "gradient": gradient,
                "int_strength": int_strength,
            }
        else:
            return {"gradient": gradient, "int_strength": int_strength}

    def KLToCurrent(self, KL: float | dict, momentum: float) -> float:
        """
        Convert the normalized strength (K value) of the magnetic field to the corresponding current.

        This method calculates the current required to produce a given normalized strength (K value)
        of the magnetic field, based on the magnet's linear and saturation fit coefficients. It accounts
        for both linear and nonlinear (saturation) behavior of the magnet.

        Args:
            KL (float): The normalized strength (K value) of the magnetic field.
                OR
            dict: A dictionary containing the K value and its gradient.
            momentum (float): The momentum of the particle beam (in MeV/c).

        Returns:
            float: The current (in amperes) required to produce the given K value.
        """
        m, I_max, f, a, I0, d, L = list(self.coefficients)
        if isinstance(KL, dict):
            if "KL" in KL:
                KL = KL["KL"]
            elif "K" in KL:
                KL = KL["K"] * L / 1000
        return self.KToCurrent(KL / (L / 1000), momentum)

    def KToCurrent(self, K: float | dict, momentum: float) -> float:
        """
        Convert the normalized strength (K value) of the magnetic field to the corresponding current.
        This method calculates the current required to produce a given normalized strength (K value)
        of the magnetic field, based on the magnet's linear and saturation fit coefficients. It accounts
        for both linear and nonlinear (saturation) behavior of the magnet.

        Args:
            K (float): The normalized strength (K value) of the magnetic field.
                OR
            dict: A dictionary containing the K value and its gradient.
            momentum (float): The momentum of the particle beam (in MeV/c).

        Returns:
            float: The current (in amperes) required to produce the given K value.
        """
        m, I_max, f, a, I0, d, L = list(self.coefficients)
        if isinstance(K, dict):
            if "K" in K:
                K = K["K"]
            elif "KL" in K:
                K = K["KL"] / (L / 1000)
            else:
                raise ValueError(f"K value not found in the dictionary {K}")
        int_strength = 1e6 * K * L * momentum / speed_of_light
        abs_str = abs(int_strength)
        linear_current = int_strength / m
        if abs(linear_current) < I_max:
            return linear_current
        elif f == 0:
            return I0 - Sqrt((abs_str - d) / a)
        else:
            p = (-6 * f * a * I0 - a**2) / (3 * f**2)
            q = (
                (2 * a**3)
                + (18 * f * a**2 * I0)
                + (27 * f**2 * (a * I0**2 + d - abs_str))
            ) / (27 * f**3)
            r = Sqrt((p / 3) ** 3)
            theta = np.arccos(-q / (2 * r))
            r_cbrt = -(r ** (1 / 3))
            t3 = 2 * r_cbrt * np.cos((theta / 3) + 4 * Pi / 3)
            return t3 - a / (3 * f)

    def __iter__(self) -> iter:
        return iter(
            [getattr(self, k) for k in ["m", "I_max", "f", "a", "I0", "d", "L"]]
        )


class MagneticElement(IgnoreExtra):
    """
    Magnetic info model.
    """

    order: int = Field(repr=False, default=-1, frozen=True)
    """Magnetic order"""

    skew: bool = False
    """Flag to indicate if the multipole is skew."""

    length: NonNegativeFloat = Field(default=0.0, alias="magnetic_length")
    """Magnetic length [m]."""

    multipoles: Multipoles = Multipoles()
    """Magnetic multipoles."""

    systematic_multipoles: Multipoles = Multipoles()
    """Systematic magnetic multipoles."""

    random_multipoles: Multipoles = Multipoles()
    """Random magnetic multipoles."""

    field_integral_coefficients: FieldIntegral | None = None  # FieldIntegral()
    """Field integral coefficients."""

    linear_saturation_coefficients: LinearSaturationFit | None = None  # LinearSaturationFit()
    """Linear saturation fit coefficients."""

    settle_time: float | None = None  # Field(alias="mag_set_max_wait_time", default=45.0)
    """
    Maximum time to wait for the magnet current to settle [s].
    #TODO move to electrical?
    """

    entrance_edge_angle: float | str = Field(default=0.0)
    """Entrance edge angle in degrees; can be "angle" which uses the bend angle."""

    exit_edge_angle: float | str = Field(default=0.0)
    """Exit edge angle in degrees; can be "angle" which uses the bend angle."""

    gap: float = Field(default=0.032)
    """Magnetic gap [m]."""

    bore: float = Field(default=0.037)
    """Magnetic bore radius [m]."""

    plane: str = Field(default="horizontal")
    """Magnetic field plane: 'horizontal' or 'vertical'."""

    width: float = Field(default=0.2)
    """Width of magnet [m]."""

    tilt: float = Field(default=0.0)
    """Tilt angle of magnet [degrees]."""

    edge_field_integral: float = Field(default=0.5)
    """Edge field integral."""

    fringe_field_coefficient: float = Field(default=0.0)
    """Fringe field coefficient."""


    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        for k in ["kl", "data"]:
            if k in data:
                self.kl = data["kl"]
                if self.skew:
                    setattr(
                        self.multipoles,
                        "K" + str(self.order) + "L",
                        Multipole(skew=self.kl, order=self.order),
                    )
                else:
                    setattr(
                        self.multipoles,
                        "K" + str(self.order) + "L",
                        Multipole(normal=self.kl, order=self.order),
                    )
        for i in range(0, 9):
            if f"k{i}l" in data:
                # print('k1l', data['k1l'])
                setattr(
                    self.multipoles,
                    f"K{i}L",
                    Multipole(normal=data[f"k{i}l"], order=i),
                )

    @field_validator("field_integral_coefficients", mode="before")
    @classmethod
    def validate_field_integral_coefficients(
        cls, v: Union[str, List, dict | None]
    ) -> FieldIntegral | None:
        if isinstance(v, str):
            return FieldIntegral(coefficients=list(map(float, v.split(","))))
        elif isinstance(v, (list, tuple)):
            return FieldIntegral(coefficients=list(v))
        elif isinstance(v, dict):
            return FieldIntegral(**v)
        elif isinstance(v, FieldIntegral):
            return v
        elif v is None:
            return None
        else:
            raise ValueError(
                "field_integral_coefficients should be a string or a list of floats"
            )

    # @debug
    def KnL(self, order: int = None) -> Union[int, float]:
        """
        Get the integrated strength (KnL) of the multipole for a given order.

        Args:
            order (int, optional): The order of the multipole. Defaults to None, which uses self.order.

        Returns:
            Union[int, float]: The integrated strength (KnL) of the multipole.
        """
        f = self.multipoles.skew if self.skew else self.multipoles.normal
        order = self.order if order is None else order
        return f(order) if order >= 0 else 0

    def Kn(self, order: int = None) -> Union[int, float]:
        """
        Get the normalized strength (Kn) of the multipole for a given order.

        Args:
            order (int, optional): The order of the multipole. Defaults to None, which uses self.order.

        Returns:
            Union[int, float]: The normalized strength (Kn) of the multipole.
        """
        return self.knl(order) / self.length

    @property
    def kl(self) -> Union[int, float]:
        return self.KnL(self.order)

    @kl.setter
    def kl(self, kl: float = 0) -> None:
        # print('kl called!', getattr(self.multipoles, 'K'+str(self.order)+'L'))
        setattr(getattr(self.multipoles, "K" + str(self.order) + "L"), "normal", kl)
        setattr(
            getattr(self.multipoles, "K" + str(self.order) + "L"), "order", self.order
        )

    @computed_field
    @property
    def half_gap(self) -> float:
        return self.gap / 2

    def gradient(self, momentum: float) -> float:
        """
        Get the magnetic field gradient for the multipole.

        Args:
            momentum (float): The momentum of the particle beam (in MeV/c).

        Returns:
            float: The magnetic field gradient.
        """
        Brho = 3.3356 * momentum / (1e9)
        return self.kl * Brho / self.length


class Dipole_Magnet(MagneticElement):
    """
    Dipole magnet with magnetic order 0.
    """

    order: int = Field(repr=False, default=0, frozen=True)
    """Magnetic order of the dipole."""

    @property
    def angle(self) -> float:
        return self.KnL(order=0)

    @angle.setter
    def angle(self, value: float) -> None:
        self.multipoles.K0L.normal = value

    @computed_field
    @property
    def rho(self) -> float:
        """
        Get the dipole bend radius -- l / theta.

        Returns
            float: The dipole bend radius
        """
        return (
            self.length / self.angle
            if self.length is not None and abs(self.angle) > 1e-9
            else 0
        )

    def field_strength(self, momentum: float) -> float:
        """
        Get the dipole magnetic field strength.

        Args:
            momentum (float): The momentum of the particle beam (in eV/c).

        Returns:
            float: The dipole magnetic field strength.
        """
        Brho = 3.3356 * momentum / (1e9)
        return self.rho * Brho / self.length


class Quadrupole_Magnet(MagneticElement):
    """
    Quadrupole with magnetic order 1.
    """

    order: int = Field(repr=False, default=1, frozen=True)
    """Magnetic order of the quadrupole."""

    @property
    def k1l(self) -> float:
        return self.kl

    @k1l.setter
    def k1l(self, value: float) -> None:
        self.kl = value


class Sextupole_Magnet(MagneticElement):
    """
    Sextupole magnet with magnetic order 2.
    """

    order: int = Field(repr=False, default=2, frozen=True)
    """Magnetic order of the sextupole."""

    @property
    def k2l(self) -> float:
        return self.kl

    @k2l.setter
    def k2l(self, value: float) -> None:
        self.kl = value

class Octupole_Magnet(MagneticElement):
    """
    Octupole magnet with magnetic order 3.
    """

    order: int = Field(repr=False, default=3, frozen=True)
    """Magnetic order of the octupole."""

    @property
    def k3l(self) -> float:
        return self.kl

    @k3l.setter
    def k3l(self, value: float) -> None:
        self.kl = value


solenoidFields = {
    "S" + str(no) + "L": (float, Field(default=0, repr=False)) for no in range(0, 13)
}
solenoidFieldsData = create_model("solenoidFieldsData", **solenoidFields)


class SolenoidFields(solenoidFieldsData):
    """Magnetic multipoles model."""

    # def __str__(self):
    #     return " ".join(
    #         [
    #             "S" + str(i) + "L=" + getattr(self, "S" + str(i) + "L").__str__() + ""
    #             for i in range(13)
    #             if abs(getattr(self, "S" + str(i) + "L")) > 0
    #         ]
    #     )

    def __repr__(self):
        return "SolenoidFields(" + self.__str__() + ")"

    @model_serializer
    def ser_model(self) -> Dict[str, Any]:
        cls = self.__class__
        return {
            k: getattr(self, k)
            for k in cls.model_fields.keys()
            if abs(getattr(self, k)) > 0
        }

    def normal(self, order: int) -> Union[int, float]:
        return getattr(self, "S" + str(order) + "L")

    def __eq__(self, other: Any) -> bool:
        return self.ser_model() == other

    def __neq__(self, other: Any) -> bool:
        return not self.__eq__(other)


class Solenoid_Magnet(IgnoreExtra):
    """
    Solenoid magnet including higher order fields.
    """

    length: NonNegativeFloat = Field(default=0.0, alias="magnetic_length")
    """Magnetic length [m]."""

    order: int = Field(repr=False, default=0, frozen=True)
    """Solenoid multipole order."""

    fields: SolenoidFields = SolenoidFields()
    """Solenoid fields."""

    systematic_fields: SolenoidFields = SolenoidFields()
    """Systematic solenoid fields."""

    random_fields: SolenoidFields = SolenoidFields()
    """Random solenoid fields."""

    field_integral_coefficients: FieldIntegral = FieldIntegral()
    """Field integral coefficients."""

    linear_saturation_coefficients: LinearSaturationFit = LinearSaturationFit()
    """Linear saturation coefficients."""

    settle_time: float = Field(alias="mag_set_max_wait_time", default=45.0)
    """Time for solenoid to settle.
    #TODO move to electrical?
    """

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        if "ks" in data:
            self.ks = data["ks"]
        elif "field_amplitude" in data:
            self.ks = data["field_amplitude"] / self.length
        # else:
        #     self.ks = 0
        # setattr(self.fields, 'S'+str(self.order)+'L', self.ks)

    @field_validator("field_integral_coefficients", mode="before")
    @classmethod
    def validate_field_integral_coefficients(cls, v: Union[str, List]) -> FieldIntegral:
        if isinstance(v, str):
            return FieldIntegral(coefficients=list(map(float, v.split(","))))
        elif isinstance(v, (list, tuple)):
            return FieldIntegral(coefficients=list(v))
        elif isinstance(v, dict):
            return FieldIntegral(**v)
        elif isinstance(v, FieldIntegral):
            return v
        else:
            raise ValueError(
                "field_integral_coefficients should be a string or a list of floats"
            )

    @property
    def field_amplitude(self) -> Union[int, float]:
        return self.ks / self.length

    @field_amplitude.setter
    def field_amplitude(self, fa: float = 0) -> None:
        self.ks = fa * self.length

    @property
    def ks(self) -> Union[int, float]:
        return getattr(self.fields, "S" + str(self.order) + "L")

    @ks.setter
    def ks(self, ks: float = 0) -> None:
        setattr(self.fields, "S" + str(self.order) + "L", ks)


class NonLinearLens_Magnet(IgnoreExtra):
    """
    Non-linear lens magnet. See `MAD-X manual`_ and `PAC2011 article`_

    .. _MAD-X manual: https://cern.ch/madx
    .. _PAC2011 article: https://proceedings.jacow.org/PAC2011/papers/wep070.pdf
    """

    length: NonNegativeFloat = Field(default=0.0, alias="magnetic_length")
    """Magnetic length of NLL [m]."""

    integrated_strength: NonNegativeFloat = Field(default=0.0, alias="knll")
    """Integrated strength of NLL."""

    dimensional_parameter: float = Field(default=0.0, alias="cnll")
    """Dimensional parameter of NLL."""

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)


class Wiggler_Magnet(IgnoreExtra):
    """
    Undulator magnet.
    """

    length: NonNegativeFloat = Field(default=0.0, alias="magnetic_length")
    """Magnetic length of wiggler [m].
    #TODO validate / check that length == period*num_periods.
    """

    strength: NonNegativeFloat = Field(default=0.0, alias="K")
    """Wiggler strength (K) parameter."""

    peak_magnetic_field: float = Field(default=0.0, alias="B")
    """Peak wiggler magnetic field [B]."""

    period: NonNegativeFloat = Field(default=0.0, alias="lambdau")
    """Wiggler period [m]."""

    num_periods: NonNegativeInt = Field(default=0, alias="nwig")
    """Number of periods in the wiggler [m]."""

    helical: bool = False
    """Flag to indicate if the wiggler is helical; False implies planar."""

    quadratic_roll_off_x: float = Field(default=0.0, alias="kx")
    """Horizontal quadratic roll-off parameter."""

    quadratic_roll_off_y: float = Field(default=0.0, alias="ky")
    """Vertical quadratic roll-off parameter."""

    transverse_gradient_x: float = Field(default=0.0, alias="gradx")
    """Horizontal transverse gradient."""

    transverse_gradient_y: float = Field(default=0.0, alias="grady")
    """Vertical transverse gradient."""

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        # setattr(self.fields, 'S'+str(self.order)+'L', self.ks)

    @property
    def normalized_strength(self) -> float:
        """
        Getter for the normalised undulator strength :math:`a_w`

        Returns
        -------
        float:
            :attr:`~strength` / :math:`\\sqrt{2}`
        """

        if not self.helical:
            return self.strength / np.sqrt(2)
        else:
            return self.strength

    @normalized_strength.setter
    def normalized_strength(self, aw: float) -> None:
        """
        Setter for the normalised undulator strength :math:`a_w`

        Parameters
        ----------
        aw: float
            :attr:`~strength` = :math:`a_w \\times \\sqrt{2}`
        """
        if not self.helical:
            self.strength = aw * np.sqrt(2)
        else:
            self.strength = aw

    @property
    def poles(self) -> int:
        """
        Number of poles, twice :attr:`~num_periods`.

        Returns
        -------
        int
            Number of poles
        """
        return int(self.num_periods * 2)

    @poles.setter
    def poles(self, value: int) -> None:
        self.num_periods = int(value / 2)


from .base import BaseElementTranslator
from nala.models.RF import WakefieldElement
from ..utils.fields import field

class WakefieldTranslator(BaseElementTranslator):
    """
    Translator class for converting a :class:`~nala.models.element.Wakefield` instance into a string or
    object that can be understood by various simulation codes.
    """

    cavity: WakefieldElement
    """Wakefield element."""

    def to_astra(self, n: int = 0, **kwargs: dict) -> str:
        """
        Writes the wakefield element string for ASTRA;
        see :func:`~_write_ASTRA`.

        Parameters
        ----------
        n: int
            Element index number
        **kwargs: dict
            Keyword args

        Returns
        -------
        str
            String representation of the element for ASTRA
        """
        self.start_write()
        return self._write_ASTRA(n=n)

    def _write_ASTRA(self, n: int = 0, **kwargs: dict) -> str:
        """
        Writes the wakefield element string for ASTRA. Each cell in a cavity gets its own &WAKE element.

        Parameters
        ----------
        n: int
            Wake index

        Returns
        -------
        str
            String representation of the element for ASTRA
        """
        field_ref_pos = self.get_field_reference_position()
        field_file_name = self.generate_field_file_name(
            self.simulation.wakefield_definition, code="astra"
        )
        efield_def = [
            "Wk_filename",
            {"value": "'" + field_file_name + "'", "default": ""},
        ]
        output = ""
        if self.simulation.wakefield_definition.field_type == "LongitudinalWake":
            waketype = "Monopole_Method_F"
        elif self.simulation.wakefield_definition.field_type == "TransverseWake":
            waketype = "Dipole_Method_F"
        else:
            waketype = "Taylor_Method_F"
        if self.simulation.scale_kick > 0:
            for n in range(n, n + int(self.cavity.n_cells)):
                output += self._write_ASTRA_dictionary(
                    dict(
                        [
                            [
                                "Wk_Type",
                                {
                                    "value": '"' + waketype + '"',
                                    "default": "'Taylor_Method_F'",
                                },
                            ],
                            efield_def,
                            ["Wk_x", {"value": self.dx, "default": 0}],
                            ["Wk_y", {"value": self.dx, "default": 0}],
                            [
                                "Wk_z",
                                {
                                    "value": self.physical.start.z
                                             + (0.5 + n - 1) * self.cavity.cell_length
                                },
                            ],
                            ["Wk_ex", {"value": self.simulation.scale_field_ex, "default": 0}],
                            ["Wk_ey", {"value": self.simulation.scale_field_ey, "default": 0}],
                            ["Wk_ez", {"value": self.simulation.scale_field_ez, "default": 1}],
                            ["Wk_hx", {"value": self.simulation.scale_field_hx, "default": 1}],
                            ["Wk_hy", {"value": self.simulation.scale_field_hy, "default": 0}],
                            ["Wk_hz", {"value": self.simulation.scale_field_hz, "default": 0}],
                            [
                                "Wk_equi_grid",
                                {"value": self.simulation.equal_grid, "default": 0.66},
                            ],
                            ["Wk_N_bin", {"value": 10, "default": 100}],
                            [
                                "Wk_ip_method",
                                {"value": self.simulation.interpolation_method, "default": 2},
                            ],
                            ["Wk_smooth", {"value": self.simulation.smooth, "default": 0.25}],
                            ["Wk_sub", {"value": self.simulation.subbins, "default": 10}],
                            [
                                "Wk_scaling",
                                {"value": 1 * self.simulation.scale_kick, "default": 1},
                            ],
                        ]
                    ),
                    n,
                )
                output += "\n"
            output += "\n"
        return output

    def to_gpt(self, Brho: float = 0.0, ccs: str="wcs", *args, **kwargs) -> str:
        """
        Write a string representation of the wakefield for GPT.

        Parameters
        ----------
        Brho: float
            Magnetic rigidity.
        ccs: str
            Name of co-ordinate system of the wakefield.

        Returns
        -------
        str
            String representation of the wakefield for GPT.
        """
        self.start_write()
        field_ref_pos = self.get_field_reference_position()
        field_file_name = self.generate_field_file_name(
            self.simulation.wakefield_definition, code="gpt"
        )
        fringe_field_coefficient = 3.0 / self.cavity.cell_length
        output = ""
        if self.simulation.scale_kick > 0:
            zcolumn = "z"
            wzcolumn = "Wz"
            wxcolumn = "Wx" if self.simulation.wakefield_definition.Wx.value is not None else ""
            wycolumn = "Wy" if self.simulation.wakefield_definition.Wy.value is not None else ""
            for n in range(self.cavity.n_cells):
                ccs_label, value_text = self.ccs.ccs_text(
                    [
                        field_ref_pos[0],
                        field_ref_pos[1],
                        field_ref_pos[2]
                        + self.cavity.coupling_cell_length
                        + n * self.cavity.cell_length,
                    ],
                    self.physical.rotation.model_dump(),
                )
                output += (
                        "wakefield"
                        + '("'
                        + self.ccs.name
                        + '", '
                        + ccs_label
                        + ", "
                        + value_text
                        + ", "
                        + str(self.cavity.cell_length)
                        + ", "
                        + str(fringe_field_coefficient)
                        + ', "'
                        + str(field_file_name)
                        + '", "'
                        + zcolumn
                        + '", "'
                        + wxcolumn
                        + '", "'
                        + wycolumn
                        + '", "'
                        + wzcolumn
                        + '");\n'
                )
        return output
